from abc import ABCMeta, abstractmethod

import numpy as np

from hazma.parameters import convolved_spectrum_fn
from hazma.theory._theory_cmb import TheoryCMB
from hazma.theory._theory_constrain import TheoryConstrain
from hazma.theory._theory_gamma_ray_limits import TheoryGammaRayLimits


class Theory(TheoryGammaRayLimits, TheoryCMB, TheoryConstrain):
    """
    Represents a sub-GeV DM theory.
    """

    __metaclass__ = ABCMeta

    @staticmethod
    @abstractmethod
    def list_annihilation_final_states():
        r"""
        Lists annihilation final states.

        Subclasses must implement this method.

        Returns
        -------
        fss : list(str)
            Possible annihilation final states.
        """
        pass

    @abstractmethod
    def annihilation_cross_section_funcs(self):
        r"""
        Gets functions to compute annihilation cross sections.

        Subclasses must implement this method.

        Returns
        -------
        sigma_fns : dict(str, (float) -> float)
            A map specifying a function for each final state which takes a
            center of mass energy and returns the corresponding annihilation
            cross section in :math:`\mathrm{MeV}^{-2}`.
        """
        pass

    def annihilation_cross_sections(self, e_cm):
        r"""
        Computes annihilation cross sections.

        Parameters
        ---------
        e_cm : float
            Center of mass energy for the annihilation in MeV.

        Returns
        -------
        sigmas : dict(str, float)
            Annihilation cross section into each final state in
            :math:`\mathrm{MeV}^{-2}` as well as the total cross section.
        """
        sigmas = {
            fs: sigma_fn(e_cm)
            for fs, sigma_fn in self.annihilation_cross_section_funcs().items()
        }
        sigmas["total"] = sum(sigmas.values())
        return sigmas

    def annihilation_branching_fractions(self, e_cm):
        r"""
        Computes annihilation branching fractions.

        Parameters
        ---------
        e_cm : float
            Center of mass energy for the annihilation in MeV.

        Returns
        -------
        bfs : dict(str, float)
            Annihilation branching fractions into each final state.
        """
        cs = self.annihilation_cross_sections(e_cm)

        if cs["total"] == 0:
            return {fs: 0.0 for fs in cs if fs != "total"}
        else:
            return {
                fs: sigma / cs["total"] for fs, sigma in cs.items() if fs != "total"
            }

    @abstractmethod
    def partial_widths(self):
        """
        Computes mediator decay widths.

        Subclasses must implement this method.

        Returns
        -------
        widths : dict(str, float)
            Mediator partial widths in MeV as the total cross decay width.
        """
        pass

    @abstractmethod
    def spectrum_funcs(self):
        r"""
        Gets a function computing the continuum gamma-ray spectrum for
        annihilations into each relevant final state.

        Subclasses must implement this method.

        Returns
        -------
        spec_fns : dict(str, (float or np.array, float) -> float)
            :math:`dN/dE_\gamma` as a function of photon energies and the
            annihilation center of mass energy for annihilation into each final
            state that produces a continuum spectrum.
        """
        pass

    def spectra(self, e_gams, e_cm):
        r"""
        Gets the contributions to the continuum gamma-ray annihilation spectrum
        for each final state.

        Parameters
        ---------
        e_gams : float or float numpy.array
            Photon energy or energies at which to compute the spectrum.
        e_cm : float
            Center of mass energy for the annihilation in MeV.

        Returns
        -------
        specs : dict(str, float)
            Contribution to :math:`dN/dE_\gamma` at the given photon energies
            and center-of-mass energy for each relevant final state. More
            specifically, this is the spectrum for annihilation into each
            channel rescaled by the corresponding branching fraction into that
            channel.
        """
        bfs = self.annihilation_branching_fractions(e_cm)
        specs = {}

        for fs, dnde_func in self.spectrum_funcs().items():
            if bfs[fs] == 0:
                specs[fs] = np.zeros_like(e_gams)
            else:
                specs[fs] = bfs[fs] * dnde_func(e_gams, e_cm)

        specs["total"] = sum(specs.values())

        return specs

    def total_spectrum(self, e_gams, e_cm):
        r"""
        Computes total continuum gamma-ray spectrum.

        Parameters
        ----------
        e_gams : float or float numpy.array
            Photon energy or energies at which to compute the spectrum.
        e_cm : float
            Annihilation center of mass energy.

        Returns
        -------
        spec : float numpy.array
            Array containing the total annihilation gamma-ray spectrum.
        """
        if hasattr(e_gams, "__len__"):
            return self.spectra(e_gams, e_cm)["total"]
        else:
            return self.spectra(np.array([e_gams]), e_cm)["total"]

    @abstractmethod
    def gamma_ray_lines(self, e_cm):
        r"""
        Gets information about annihilation into gamma-ray lines.

        Subclasses must implement this method.

        Parameters
        ----------
        e_cm : float
            Annihilation center of mass energy.

        Returns
        -------
        lines : dict(str, dict(str, float))
            For each final state containing a monochromatic photon, gives the
            energy of the photon and branching fraction into that final state.
        """
        pass

    def total_conv_spectrum_fn(
        self, e_gam_min, e_gam_max, e_cm, energy_res, n_pts=1000
    ):
        r"""
        Computes the total gamma-ray spectrum convolved with an energy
        resolution function.

        Parameters
        ----------
        e_min : float
            Lower bound of energy range over which to perform convolution.
        e_max : float
            Upper bound of energy range over which to perform convolution.
        e_cm : float
            Center of mass energy for DM annihilation.
        energy_res : float -> float
            The detector's energy resolution (Delta E / E) as a function of
            photon energy in MeV.
        n_pts : float
            Number of points to use to create resulting interpolating function.
            More points gives higher accuracy at the cost of computing time,
            but is necessary if the continuum spectrum contains very sharp
            features.

        Returns
        -------
        dnde_conv : InterpolatedUnivariateSpline
            An interpolator giving the DM annihilation spectrum smeared by the
            energy resolution function. Using photon energies outside the range
            [e_min, e_max] will produce a ``bounds_error``.
        """
        return convolved_spectrum_fn(
            e_gam_min,
            e_gam_max,
            e_cm,
            energy_res,
            self.total_spectrum,
            self.gamma_ray_lines,
            n_pts,
        )

    @abstractmethod
    def positron_spectrum_funcs(self):
        r"""
        Gets a function computing the continuum positron spectrum for
        annihilations into each relevant final state.

        Subclasses must implement this method.

        Returns
        -------
        spec_fns : dict(str, (float or np.array, float) -> float)
            :math:`dN/dE_{e^+}` as a function of positron energies and the
            annihilation center of mass energy for annihilation into each final
            state that produces a continuum spectrum.
        """
        pass

    def positron_spectra(self, e_ps, e_cm):
        r"""
        Gets the contributions to the continuum positron annihilation spectrum
        for each final state.

        Parameters
        ---------
        e_ps : float or float numpy.array
            Positron energy or energies at which to compute the spectrum.
        e_cm : float
            Center of mass energy for the annihilation in MeV.

        Returns
        -------
        specs : dict(str, float)
            Contribution to :math:`dN/dE_\gamma` at the given positron energies
            and center-of-mass energy for each relevant final state. More
            specifically, this is the spectrum for annihilation into each
            channel rescaled by the corresponding branching fraction into that
            channel.
        """
        bfs = self.annihilation_branching_fractions(e_cm)
        dndes_pos = {}

        for fs, dnde_pos_func in self.positron_spectrum_funcs().items():
            if bfs[fs] == 0:
                dndes_pos[fs] = np.zeros_like(e_ps)
            else:
                dndes_pos[fs] = bfs[fs] * dnde_pos_func(e_ps, e_cm)

        dndes_pos["total"] = sum(dndes_pos.values())

        return dndes_pos

    def total_positron_spectrum(self, e_ps, e_cm):
        r"""
        Computes the total positron ray spectrum.

        Parameters
        ----------
        e_ps : float or float numpy.array
            Positron energy or energies at which to compute the spectrum.
        e_cm : float
            Annihilation center of mass energy.

        Returns
        -------
        spec : float numpy.array
            Array containing the total annihilation positron spectrum.
        """
        if hasattr(e_ps, "__len__"):
            return self.positron_spectra(e_ps, e_cm)["total"]
        else:
            return self.positron_spectra(np.array([e_ps]), e_cm)["total"]

    @abstractmethod
    def positron_lines(self, e_cm):
        r"""
        Gets information about annihilation into monochromatic positrons.

        Subclasses must implement this method.

        Parameters
        ----------
        e_cm : float
            Annihilation center of mass energy.

        Returns
        -------
        lines : dict(str, dict(str, float))
            For each final state containing a monochromatic positron, gives the
            energy of the positron and branching fraction into that final
            state.
        """
        pass

    def total_conv_positron_spectrum_fn(
        self, e_p_min, e_p_max, e_cm, energy_res, n_pts=1000
    ):
        r"""
        Computes the total positron spectrum convolved with an energy
        resolution function.

        Parameters
        ----------
        e_min : float
            Lower bound of energy range over which to perform convolution.
        e_max : float
            Upper bound of energy range over which to perform convolution.
        e_cm : float
            Center of mass energy for DM annihilation.
        energy_res : float -> float
            The detector's energy resolution (Delta E / E) as a function of
            positron energy in MeV.
        n_pts : float
            Number of points to use to create resulting interpolating function.
            More points gives higher accuracy at the cost of computing time,
            but is necessary if the continuum spectrum contains very sharp
            features.

        Returns
        -------
        dnde_conv : InterpolatedUnivariateSpline
            An interpolator giving the DM annihilation spectrum smeared by the
            energy resolution function. Using positron energies outside the
            range [e_min, e_max] will produce a ``bounds_error``.
        """
        return convolved_spectrum_fn(
            e_p_min,
            e_p_max,
            e_cm,
            energy_res,
            self.total_positron_spectrum,
            self.positron_lines,
            n_pts,
        )

    @abstractmethod
    def constraints(self):
        r"""
        Get a dictionary of all available constraints.

        Subclasses must implement this method.

        Notes
        -----
        Each key in the dictionary is the name of a constraint. Each value is a
        function that is positive when the constraint is satisfied and negative
        when it is not.
        """
        raise NotImplementedError()



from hazma.decay_helper_functions.decay_muon cimport CSpectrum as muspec
from hazma.decay_helper_functions.decay_muon cimport CSpectrumPoint as muspecpt
import numpy as np
cimport numpy as np
from scipy.integrate import quad
from libc.math cimport exp, log, M_PI, log10, sqrt, abs, pow
import cython
include "parameters.pxd"

import warnings

cdef double eng_gam_max_mu_rf = (MASS_MU**2.0 - MASS_E**2.0) / (2.0 * MASS_MU)
cdef double eng_mu_pi_rf = (MASS_PI**2.0 + MASS_MU**2.0) / (2.0 * MASS_PI)
cdef double fpi = DECAY_CONST_PI / sqrt(2.)
cdef double mpi = MASS_PI
cdef double me = MASS_E
cdef double mmu = MASS_MU

cdef np.ndarray eng_gams_mu
cdef np.ndarray mu_spec

eng_gams_mu = np.logspace(-5.5, 3.0, num=10000, dtype=np.float64)
mu_spec = muspec(eng_gams_mu, eng_mu_pi_rf)


cdef double muon_spectrum(double eng_gam):
    """
    Returns the interpolated muon spectrum.
    """
    return np.interp(eng_gam, eng_gams_mu, mu_spec)


@cython.cdivision(True)
cdef double dnde_pi_to_lnug(double egam, double ml):
    """
    Returns dnde from pi-> l nu g.
    """
    cdef double r = pow(ml / mpi, 2.0)
    if 0.0 <= egam and egam <= mpi * (1. - r) / 2. :
        return (ALPHA_EM *
            (-2 * (2 * egam + mpi * (-1 + r)) *
             (3 * fpi**2 * (2 * egam - mpi) * mpi * r *
              ((egam - mpi)**2 +
               (2 * egam - mpi) * mpi * r) - 3 * egam**2 * fpi *
              (2 * egam - mpi) * mpi * r *
              (4 * A_PI * egam - A_PI * mpi *
              (1 + r) - 2 * egam * V_PI) + egam**4 *
              (2 * egam + mpi * (-1 + r)) * (4 * egam - mpi * (2 + r)) *
              (A_PI**2 + V_PI**2)) + 3 * fpi * mpi * (-2 * egam + mpi)**2 * r *
             (2 * egam**2 * fpi + 2 * egam * fpi * mpi *
              (-1 + r) - 4 * A_PI * egam**2 * (egam - mpi * r) - fpi * mpi**2 *
              (-1 + r**2) + 4 * egam**3 * V_PI) *
             (log(-2 * egam + mpi) - log(mpi * r)))) / \
        (6. * egam * fpi**2 * (
            1 - (2 * egam) / mpi)**2 * mpi**4 * M_PI * (-1 + r)**2 * r)
    else :
        return 0.0



@cython.cdivision(True)
cdef double gamma(double eng, double mass):
    """
    Returns special relativity boost factor gamma.

    Keyword arguments::
        eng -- energy of particle.
        mass -- mass of particle.
    """
    return eng / mass


@cython.cdivision(True)
cdef double beta(double eng, double mass):
    """
    Returns velocity in natural units.

    Keyword arguments::
        eng -- energy of particle.
        mass -- mass of particle.
    """
    return sqrt(1.0 - (mass / eng)**2.0)


cdef double eng_gam_max(double eng_pi):
    """
    Returns the maximum allowed gamma ray energy from a charged pion decay.

    Keyword arguments::
        eng_pi: Energy of pion in laboratory frame.

    More details:
        This is computed using the fact that in the mu restframe, the
        Maximum allowed value is
            eng_gam_max_mu_rf = (pow(mass_mu,2.0) - pow(mass_e,2.0))
                            / (2.0 * mass_mu).
        Then, boosting into the pion rest frame, then to the mu rest
        frame, we get the maximum allowed energy in the lab frame.
    """
    cdef double beta_pi = beta(eng_pi, MASS_PI)
    cdef double gamma_pi = gamma(eng_pi, MASS_PI)

    cdef double beta_mu = beta(eng_mu_pi_rf, MASS_MU)
    cdef double gamma_mu = gamma(eng_mu_pi_rf, MASS_MU)

    return eng_gam_max_mu_rf * gamma_pi * gamma_mu * \
        (1.0 + beta_pi) * (1.0 + beta_mu)


@cython.cdivision(True)
cdef double integrand(double cl, double eng_gam, double eng_pi, str mode):
    """
    Returns the integrand of the differential radiative decay spectrum for
    the charged pion.

    Keyword arguments::
        cl: Angle of photon w.r.t. charged pion in lab frame.
        engPi: Energy of photon in laboratory frame.
        engPi: Energy of pion in laboratory frame.
    """
    cdef double beta_pi = beta(eng_pi, MASS_PI)
    cdef double gamma_pi = gamma(eng_pi, MASS_PI)

    cdef double beta_mu = beta(eng_mu_pi_rf, MASS_MU)
    cdef double gamma_mu = gamma(eng_mu_pi_rf, MASS_MU)

    cdef double eng_gam_pi_rF = eng_gam * gamma_pi * (1.0 - beta_pi * cl)

    cdef double jac = 1. / (2.0 * gamma_pi * abs(1.0 - beta_pi * cl))

    cdef double dnde_munu = 0.
    cdef double dnde_munug = 0.
    cdef double dnde_enug = 0.

    __eng_gam_max_pi_rf = eng_gam_max_mu_rf * gamma_mu * (1.0 + beta_mu)

    if 0. < eng_gam_pi_rF and eng_gam_pi_rF < __eng_gam_max_pi_rf:
        dnde_munu = BR_PI_TO_MUNU * jac * muon_spectrum(eng_gam_pi_rF)

    dnde_munug = BR_PI_TO_MUNU * jac * dnde_pi_to_lnug(eng_gam_pi_rF, mmu)
    dnde_enug = 0.000123 * jac * dnde_pi_to_lnug(eng_gam_pi_rF, me)

    if mode == "total":
        return dnde_munu + dnde_munug + dnde_enug
    if mode == "munu":
        return dnde_munu
    if mode == "munug":
        return dnde_munug
    if mode == "enug":
        return dnde_enug

cdef double CSpectrumPoint(double eng_gam, double eng_pi, str mode):
    """
    Returns the radiative spectrum value from charged pion given a gamma
    ray energy eng_gam and charged pion energy eng_pi. When the
    ChargedPion object is instatiated, an interplating function for the
    mu spectrum is computed.

    Keyword arguments::
        eng_gam: Energy of photon is laboratory frame.
        eng_pi: Energy of charged pion in laboratory frame.
    """
    message = 'Energy of pion cannot be less than the pion mass. Returning 0.'
    if eng_pi < MASS_PI:
        return 0.0

    return quad(integrand, -1.0, 1.0, points=[-1.0, 1.0], \
                  args=(eng_gam, eng_pi, mode), epsabs=10**-10., \
                  epsrel=10**-5.)[0]


@cython.boundscheck(True)
@cython.wraparound(False)
cdef np.ndarray CSpectrum(np.ndarray eng_gams, double eng_pi, str mode):
    """
    Returns the radiative spectrum dNde from charged pion given a gamma
    ray energies eng_gams and charged pion energy eng_pi.
    When the ChargedPion object is instatiated, an interplating function
    for the mu spectrum is computed.

    Keyword arguments::
        eng_gams: Gamma ray energies to evaluate spectrum.
        eng_pi: Energy of charged pion in laboratory frame.
    """
    cdef int numpts = len(eng_gams)

    cdef np.ndarray spec = np.zeros(numpts, dtype=np.float64)

    cdef int i = 0

    for i in range(numpts):
        spec[i] = CSpectrumPoint(eng_gams[i], eng_pi, mode)

    return spec


def SpectrumPoint(double eng_gam, double eng_pi, str mode):
    """
    Returns the radiative spectrum value from charged pion given a gamma
    ray energy eng_gam and charged pion energy eng_pi. When the
    ChargedPion object is instatiated, an interplating function for the
    mu spectrum is computed.

    Keyword arguments::
        eng_gam: Energy of photon is laboratory frame.
        eng_pi: Energy of charged pion in laboratory frame.
    """
    return CSpectrumPoint(eng_gam, eng_pi, mode)


@cython.boundscheck(True)
@cython.wraparound(False)
def Spectrum(np.ndarray eng_gams, double eng_pi, str mode):
    """
    Returns the radiative spectrum dNde from charged pion given a gamma
    ray energies eng_gams and charged pion energy eng_pi.
    When the ChargedPion object is instatiated, an interplating function
    for the mu spectrum is computed.

    Keyword arguments::
        eng_gams: Gamma ray energies to evaluate spectrum.
        eng_pi: Energy of charged pion in laboratory frame.
    """
    return CSpectrum(eng_gams, eng_pi, mode)



import numpy as np


class VectorMediatorPositronSpectra:
    def positron_spectra(self, eng_ps, cme):
        """Computes total continuum positron spectrum.
        """

        total = np.zeros(len(eng_ps))

        return {"total": total}

    def positron_lines(self, cme):
        return {}



from hazma.positron_helper_functions.positron_muon \
    cimport CSpectrum as mu_spec
from hazma.positron_helper_functions.positron_charged_pion \
    cimport CSpectrum as cp_spec

import cython
import numpy as np
cimport numpy as np
from scipy.integrate import quad

from libc.math cimport M_PI, sqrt, pow, log10

include "../decay_helper_functions/parameters.pxd"

cdef double mmu = MASS_MU
cdef double me = MASS_E
cdef double mpi = MASS_PI
cdef double qe = sqrt(4. * M_PI * ALPHA_EM)
ctypedef np.ndarray ndarray

# Cached values of the mediator mass and partial widths. If the mass of the
# mediator or partial widths change, the spectra need to be recomputed.
cdef double cache_mv = -1.0;
cdef np.ndarray cache_pws = np.array([-1.0, -1.0, -1.0])

# Set up arrays for the interpolating positron spectra for the charged pion
# and the muon.
cdef int n_interp_pts = 500
cdef np.ndarray __e_ps = np.zeros((n_interp_pts,), dtype=np.float64)
cdef np.ndarray __spec_cp = np.zeros((n_interp_pts,), dtype=np.float64)
cdef np.ndarray __spec_mu = np.zeros((n_interp_pts,), dtype=np.float64)

cdef int __recompute_rf_spectra(double mv, np.ndarray[double] pws):
    """
    Determine if we need to recompute the positron spectra for the muon and
    pion spectra in the vector rest frame.

    Parameters
    ----------
    mv: double
        Mass of the vector mediator
    pws: np.ndarray
        Relevant partial widths: pws[0] = pw_ee, pws[1] = pw_mumu and
        pws[2] = pw_pipi

    Returns
    -------
    should_recompute:
        True if we need to recompute the spectra.
    """
    global cache_mv
    global cache_pws

    if mv != cache_mv or pws[0] != cache_pws[0] or pws[1] != cache_pws[1] or \
            pws[2] != cache_pws[2]:
        return 1
    return 0

cdef void __set_spectra(double mv):
    """
    Set interpolating functions for charged pion and muon positron spectra to
    speed up functions calls during integration.

    Parameters
    ----------
    mv: double
        Mass of the vector mediator

    """
    global __e_ps
    global __spec_cp
    global __spec_mu

    __e_ps = np.logspace(log10(me), log10(mv / 2.), num=n_interp_pts)
    __spec_cp = cp_spec(__e_ps, mv / 2.)
    __spec_mu = mu_spec(__e_ps, mv / 2.)

cdef double __interp_spec(double eng_p, str fs):
    """
    Return the positron spectrum from the decay of the vector mediator into
    either a charged pion or muon for a given electron/positron energy.

    Parameters
    ----------
    eng_p: double
        Energy of the positron/electron
    fs: str
        String specifying the final state: 'cp' for charged pion and 'mu' for
        the muon.

    Returns
    -------
    spectrum: double
        The positron spectrum value from the decay of the vector mediator into
        pions or muons.

    """
    if fs == "cp":
        return np.interp(eng_p, __e_ps, __spec_cp)
    if fs == "mu":
        return np.interp(eng_p, __e_ps, __spec_mu)
    else:
        return 0.0

@cython.cdivision(True)
@cython.boundscheck(True)
@cython.wraparound(False)
cdef double __integrand(double cl, double eng_p, double eng_v,
                        double mv, np.ndarray[double] pws, str fs):
    """
    Integrand of the boost integral.

    Parameters
    ----------
    cl : float
        Angle the final state particle make with respect to the z-axis.
    eng_p : float
        Gamma-ray energy to evaluate spectrum at.
    eng_v : float
        Energy of the vector mediator.
    mv : double
        Vector mediator mass.
    pws: np.ndarray
        Relevant partial widths: pws[0] = pw_ee, pws[1] = pw_mumu and
        pws[2] = pw_pipi
    fs: str
        String specifying the final state to compute spectrum for.

    Returns
    -------
    integrand : float
        The value of the boost integral.
    """
    if eng_p < me:
        return 0.0

    cdef double pwmumu = pws[1]
    cdef double pwpipi = pws[2]

    cdef double p = sqrt(eng_p * eng_p - me * me)
    cdef double gamma = eng_v / mv
    cdef double beta = sqrt(1. - pow(mv / eng_v, 2))
    cdef double eng_p_vrf = gamma * (eng_p - p * beta * cl)
    cdef double jac = p / (2. * sqrt((1 + pow(beta * cl, 2)) * eng_p * eng_p -
                                     (1 + beta * beta * (-1 + cl * cl)) *
                                     me * me -
                                     2 * beta * cl * eng_p * p) * gamma)

    cdef double dnde_cp = 0.0
    cdef double dnde_mu = 0.0

    if fs == "total":
        dnde_cp = 2. * pwpipi * __interp_spec(eng_p_vrf, "cp")
        dnde_mu = 2. * pwmumu * __interp_spec(eng_p_vrf, "mu")
        return jac * (dnde_cp + dnde_mu)
    if fs == "pi pi":
        dnde_cp = 2. * pwpipi * __interp_spec(eng_p_vrf, "cp")
        return jac * (dnde_cp + dnde_mu)
    if fs == "mu mu":
        dnde_mu = 2. * pwmumu * __interp_spec(eng_p_vrf, "mu")
        return jac * (dnde_cp + dnde_mu)
    if fs == "e e":
        return 0.0

@cython.boundscheck(True)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double __dnde_decay_v(double eng_p, double eng_v, double mv,
                           np.ndarray[double] pws, str fs):
    """
    Un-vectorized dnde_decay_s

    Compute the gamma ray spectrum from the decay of the scalar mediator.

    Parameters
    ----------
    eng_p : double
        Positron/electron energy to evaluate spectrum at.
    eng_v : float
        Energy of the vector mediator.
    mv : double
        Mass of the vector mediator.
    pws: np.ndarray[double]
        Array of the relevant partial widths: pws[0] = pw_ee,
        pws[1] = pw_mumu and pws[2] = pw_pipi

    Returns
    -------
    dnde : float or array-like
        Values of dnde at positron/electron energy `eng_p`.
    """
    cdef double lines_contrib = 0.0
    cdef double beta = sqrt(1. - pow(mv / eng_v, 2.))
    cdef double gamma = eng_v / mv

    if eng_v < mv:
        return 0.

    cdef double r = sqrt(1.0 - 4.0 * me * me / (mv * mv))
    cdef double eplus = eng_v * (1. + r * beta) / 2.0
    cdef double eminus = eng_v * (1. - r * beta) / 2.0
    cdef double result = 0.0

    if eminus <= eng_p <= eplus:
        lines_contrib = pws[0] * 1. / (eng_v * beta)

    if fs == "e e":
        return lines_contrib

    if fs == "total" or fs == "pi pi" or fs == "mu mu":
        result = quad(__integrand, -1.0, 1.0, points=[-1.0, 1.0],
                      args=(eng_p, eng_v, mv, pws, fs), epsabs=1e-10,
                      epsrel=1e-5)[0]

        return result + lines_contrib

    return result

@cython.boundscheck(True)
@cython.wraparound(False)
def dnde_decay_v_pt(double eng_p, double eng_v, double mv,
                    np.ndarray[double] pws, str fs):
    """
    Compute the gamma ray spectrum from the decay of the scalar mediator.

    Parameters
    ----------
    eng_p : float
        Positron energy to evaluate spectrum at.
    eng_v : float
        Energy of the scalar mediator.
    mv : double
        Mass of the scalar mediator.
    pws: np.ndarray[double]
        Array of the relevant partial widths: pws[0] = pw_ee,
        pws[1] = pw_mumu and pws[2] = pw_pipi
    fs: str
        String specifying the final state to compute spectrum for.

    Returns
    -------
    dnde : float or array-like
        Value of dnde at positron energy `eng_p`.
    """
    if __recompute_rf_spectra(mv, pws) == 1:
        __set_spectra(mv)
    return __dnde_decay_v(eng_p, eng_v, mv, pws, fs)

@cython.boundscheck(True)
@cython.wraparound(False)
def dnde_decay_v(np.ndarray[double] eng_ps, double eng_v, double mv,
                 np.ndarray[double] pws, str fs):
    """
    Compute the gamma ray spectrum from the decay of the scalar mediator.

    Parameters
    ----------
    eng_ps : float
        Positron energy to evaluate spectrum at.
    eng_v : float
        Energy of the scalar mediator.
    mv : double
        Mass of the scalar mediator.
    pws: np.ndarray[double]
        Array of the relevant partial widths: pws[0] = pw_ee,
        pws[1] = pw_mumu and pws[2] = pw_pipi
    fs: str
        String specifying the final state to compute spectrum for.

    Returns
    -------
    dnde : float or array-like
        Value of dnde at positron energy `eng_p`.
    """
    if __recompute_rf_spectra(mv, pws) == 1:
        __set_spectra(mv)
    cdef int num_pts = len(eng_ps)
    cdef int i

    spec = np.zeros(num_pts, dtype=np.float64)

    for i in range(num_pts):
        spec[i] = __dnde_decay_v(eng_ps[i], eng_v, mv, pws, fs)

    return spec



# Hazma errors


class HazmaError(Exception):
    """Base class for exceptions in Hazma."""

    pass


class NegativeSquaredMatrixElementError(HazmaError):
    """
    Exception raised when a negative squared matrix element
    is encountered.
    """

    def __init__(self, message=None):
        super(NegativeSquaredMatrixElementError, self).__init__(message)
        self.default_message = "Negative squared matrix element encountered."

    def __str__(self):
        return self.default_message


class RamboCMETooSmall(HazmaError):
    """
    Exception raised when RAMBO is called with a CME less than the sum of final
    state masses.
    """

    def __init__(self, message=None):
        super(RamboCMETooSmall, self).__init__(message)
        self.default_message = (
            "Center of mass energy is less than sum of " + "the final state masses"
        )

    def __str__(self):
        return self.default_message


# Hazma warnings


class HazmaWarning(UserWarning):
    """Base class for exceptions in Hazma."""

    pass


class NegativeSquaredMatrixElementWarning(HazmaWarning):
    """
    Warning raised when a negative squared matrix element
    is encountered.
    """

    def __init__(self, message=None):
        super(NegativeSquaredMatrixElementWarning, self).__init__(message)
        self.default_message = (
            "Negative squared matrix element" + "encountered. Using zero instead."
        )

    def __str__(self):
        return self.default_message


class PreAlphaWarning(HazmaWarning):
    """
    Warning raised when user tries to access feature which is pre-alpha
    """

    def __init__(self, message=None):
        super(PreAlphaWarning, self).__init__(message)
        self.default_message = "Accessing pre-alpha feature."

    def __str__(self):
        return self.default_message



import numpy as np
cimport numpy as np
import cython
include "parameters.pxd"

cdef double CSpectrumPoint(double eng_gam, double eng_e):
    """
    Compute dN_{\gamma}/dE_{\gamma} from mu -> e nu nu gamma in the
    laborartory frame.

    Keyword arguments::
        eng_gam (float) -- Gamma ray energy in laboratory frame.
        eng_mu (float) -- Muon energy in laboratory frame.
    """
    if eng_e < MASS_E:
        raise ValueError('Energy of electron cannot be less than the electron mass.')

    cdef double result = 0.0

    return result


@cython.cdivision(True)
cdef np.ndarray CSpectrum(np.ndarray eng_gams, double eng_e):
    """
    Compute dN/dE from mu -> e nu nu gamma in the laborartory frame.

    Paramaters
    ----------
    eng_gams : np.ndarray
        List of gamma ray energies in laboratory frame.
    eng_mu : float
        Muon energy in laboratory frame.

    Returns
    -------
    spec : np.ndarray
        List of gamma ray spectrum values, dNdE, evaluated at `eng_gams`
        given muon energy `eng_mu`.
    """
    if eng_e < MASS_E:
        raise ValueError('Energy of electron cannot be less than the electron mass.')
    cdef np.ndarray spec = np.zeros(len(eng_gams), dtype=np.float64)

    return spec



import numpy as np
from scipy import optimize
from scipy.interpolate import InterpolatedUnivariateSpline


class TheoryGammaRayLimits:
    def __get_product_spline(self, f1, f2, grid, k=1, ext="raise"):
        """Returns a spline representing the product of two functions.

        Parameters
        ----------
        f1, f2 : float -> float
            The two functions to multiply.
        grid : numpy.array
            The grid used to create the product spline.
        k : int
            Degree of returned spline, 1 <= k <= 5.
        ext : string or int
            Extrapolation method. See documentation for
            `InterpolatedUnivariateSpline`.

        Returns
        -------
        spl : InterpolatedUnivariateSpline
            A degree k spline created using grid for the x array and
            f1(grid)*f2(grid) for the y array, with the specified extrapolation
            method.
        """
        return InterpolatedUnivariateSpline(grid, f1(grid) * f2(grid), k=k, ext=ext)

    def binned_limit(self, measurement, n_sigma=2.0):
        r"""
        Determines the limit on :math:`<sigma v>` from gamma-ray data.

        We define a signal to be in conflict with the measured flux for bin
        :math:`i` for an experiment if

        .. math::

            \Phi_{\chi}^{(i)} > n_{\sigma} \sigma^{(i)} + \Phi^{(i)},

        where :math:`\Phi_\chi^{(i)}` is the integrated flux due to DM
        annihilations for the bin, :math:`\Phi^{(i)}` is the measured flux in
        the bin, :math:`\sigma^{(i)}` is size of the upper error bar for the
        bin and :math:`n_{\sigma} = 2` is the significance. The overall limit on
        :math:`\langle\sigma v\rangle` is computed by minimizing over the
        limits determined for each bin.

        Parameters
        ----------
        measurement : FluxMeasurement
            Information about the flux measurement and target.
        n_sigma : float
            See the notes for this function.

        Returns
        -------
        <sigma v>_tot : float
            Largest allowed thermally averaged total cross section in cm^3 / s

        """
        # Factor to convert dN/dE to Phi. Factor of 2 comes from DM not being
        # self-conjugate.
        dm_flux_factor = (
            measurement.target.J
            * measurement.target.dOmega
            / (2.0 * 4.0 * np.pi * 2.0 * self.mx ** 2)
        )

        # TODO: this should depend on the target!
        vx = 1e-3  # v_x = Milky Way velocity dispersion
        e_cm = 2.0 * self.mx * (1.0 + 0.5 * vx ** 2)

        # Convolve the spectrum with the detector's spectral resolution
        e_min, e_max = measurement.e_lows[0], measurement.e_highs[-1]
        dnde_conv = self.total_conv_spectrum_fn(
            e_min, e_max, e_cm, measurement.energy_res
        )

        def bin_lim(e_low, e_high, phi, sigma):
            """Subroutine to compute limit in a single bin."""
            # Compute integrated flux from DM annihilation
            phi_dm = dm_flux_factor * dnde_conv.integral(e_low, e_high)

            # If flux is finite and nonzero, set a limit
            if not np.isnan(phi_dm) and phi_dm > 0:
                # Compute maximum allow integrated flux
                phi_max = (
                    measurement.target.dOmega
                    * (e_high - e_low)
                    * (n_sigma * sigma + phi)
                )

                assert phi_max > 0

                # Find limit on <sigma v>
                return phi_max / phi_dm
            else:
                return np.inf

        # Compute limits for each bin
        sv_lims = []

        for (e_low, e_high, phi, sigma) in zip(
            measurement.e_lows,
            measurement.e_highs,
            measurement.fluxes,
            measurement.upper_errors,
        ):
            sv_lims.append(bin_lim(e_low, e_high, phi, sigma))

        # Return the most stringent limit
        return np.min(sv_lims)

    def __f_jac_lim(self, e_ab, integrand_S, integrand_B):
        """Computes signal-to-noise ratio and Jacobian for an energy window.

        Notes
        -----
        This is only used by :func:`unbinned_limit`.

        Parameters
        ----------
        e_ab : (float, float)
            Lower and upper boundaries for energy window.
        integrand_S : InterpolatedUnivariateSpline
            The part of the DM spectrum times effective area dependent on
            photon energy.
        integrand_B : InterpolatedUnivariateSpline
            The part of the background spectrum times effective area dependent
            on photon energy.

        Returns
        -------
        f, jac : float, float
            Negative one times the signal-to-noise ratio (up to factors
            independent of the energy window) and f's Jacobian with respect to
            the window boundaries.
        """
        e_a, e_b = e_ab.min(), e_ab.max()

        if e_a == e_b:
            return 0.0, 0.0
        else:
            I_S_val = integrand_S.integral(e_a, e_b)
            I_B_val = integrand_B.integral(e_a, e_b)

            # Jacobian
            df_de_a = (
                1.0
                / np.sqrt(I_B_val)
                * (integrand_S(e_a) - 0.5 * I_S_val / I_B_val * integrand_B(e_a))
            )
            df_de_b = (
                -1.0
                / np.sqrt(I_B_val)
                * (integrand_S(e_b) - 0.5 * I_S_val / I_B_val * integrand_B(e_b))
            )
            jac_val = np.array([df_de_a, df_de_b]).T

            return -I_S_val / np.sqrt(I_B_val), jac_val

    def unbinned_limit(
        self,
        A_eff,
        energy_res,
        T_obs,
        target_params,
        bg_model,
        n_sigma=5.0,
        debug_msgs=False,
    ):
        r"""
        Computes smallest-detectable value of <sigma v> for given target and
        experiment parameters.

        We define a signal to be detectable if

        .. math::

            N_S / \sqrt{N_B} \geq n_{\sigma},

        where :math:`N_S` and :math:`N_B` are the number of signal and
        background photons in the energy window of interest and
        :math:`n_\sigma` is the significance in number of standard deviations.
        Note that :math:`N_S \propto \langle \sigma v \rangle`. While the
        photon count statistics are properly taken to be Poissonian and using a
        confidence interval would be more rigorous, this procedure provides a
        good estimate and is simple to compute. The energy window is chosen to
        maximize N_S/sqrt(N_B).

        Parameters
        ----------
        A_eff : float -> float
            Effective area of experiment in cm^2 as a function of photon
            energy.
        energy_res : float -> float
            The detector's energy resolution (:math:`\Delta E / E`) as a
            function of photon energy in MeV.
        T_obs : float
            Experiment's observation time in s
        target_params : TargetParams
            Object containing information about the observation target.
        bg_model : BackgroundModel
            Object representing a gamma ray background model.
        n_sigma : float
            Number of standard deviations the signal must be above the
            background to be considered detectable
        debug_msgs : bool
            If True, the energy window found by the optimizer will be printed.

        Returns
        -------
        <sigma v> : float
            Smallest-detectable thermally averaged total cross section in units
            of cm^3 / s.
        """
        # TODO: this should depend on the target!
        vx = 1e-3  # v_x = Milky Way velocity dispersion
        e_cm = 2.0 * self.mx * (1.0 + 0.5 * vx ** 2)

        # Convolve the spectrum with the detector's spectral resolution
        e_min, e_max = A_eff.x[[0, -1]]
        dnde_conv = self.total_conv_spectrum_fn(e_min, e_max, e_cm, energy_res)

        # Energy at which spectrum peaks.
        e_dnde_max = optimize.minimize(
            dnde_conv, 0.5 * (e_min + e_max), bounds=[[e_min, e_max]]
        ).x[0]

        # If there's a peak, include it in the initial energy window.
        if np.isclose(e_dnde_max, e_min, atol=0, rtol=1e-8) or np.isclose(
            e_dnde_max, e_max, atol=0, rtol=1e-8
        ):  # no peak
            e_a_0 = 10.0 ** (0.15 * np.log10(e_max) + 0.85 * np.log10(e_min))
            e_b_0 = 10.0 ** (0.85 * np.log10(e_max) + 0.15 * np.log10(e_min))
        else:
            e_a_0 = 10.0 ** (0.5 * (np.log10(e_dnde_max) + np.log10(e_min)))
            e_b_0 = 10.0 ** (0.5 * (np.log10(e_max) + np.log10(e_dnde_max)))

        # Integrating these gives the number of signal and background photons
        integrand_S = self.__get_product_spline(dnde_conv, A_eff, dnde_conv.get_knots())
        integrand_B = self.__get_product_spline(
            bg_model.dPhi_dEdOmega, A_eff, dnde_conv.get_knots()
        )

        # Optimize upper and lower bounds for energy window
        limit_obj = optimize.minimize(
            self.__f_jac_lim,
            [e_a_0, e_b_0],
            args=(integrand_S, integrand_B),
            bounds=2 * [[1.001 * e_min, e_max]],
            constraints=({"type": "ineq", "fun": lambda x: x[1] - x[0]}),
            jac=True,
            options={"ftol": 1e-7, "eps": 1e-10},
            method="SLSQP",
        )

        if debug_msgs:
            print("\te_a, e_b = ".format(limit_obj.x))

        # Insert appropriate prefactors to convert result to <sigma v>_tot. The
        # factor of 2 is from the DM not being self-conjugate.
        prefactor = (
            2.0
            * 4.0
            * np.pi
            * 2.0
            * self.mx ** 2
            / (np.sqrt(T_obs * target_params.dOmega) * target_params.J)
        )

        assert -limit_obj.fun >= 0

        return prefactor * n_sigma / (-limit_obj.fun)



import numpy as np
cimport numpy as np

cdef np.ndarray __e_gams_total
cdef np.ndarray __e_gams_00
cdef np.ndarray __e_gams_pm
cdef np.ndarray __e_gams_pmg

cdef np.ndarray __spec_total
cdef np.ndarray __spec_00
cdef np.ndarray __spec_pm
cdef np.ndarray __spec_pmg

cdef double __interp_spec(double, str)
cdef double __integrand(double, double, double, str)

cdef double CSpectrumPoint(double, double, str)
cdef np.ndarray CSpectrum(np.ndarray[np.float64_t, ndim=1], double, str)



from hazma.theory import Theory

from hazma.parameters import up_quark_mass as muq
from hazma.parameters import down_quark_mass as mdq
from hazma.parameters import strange_quark_mass as msq
from hazma.parameters import fpi, b0, vh

from hazma.scalar_mediator._scalar_mediator_constraints import ScalarMediatorConstraints
from hazma.scalar_mediator._scalar_mediator_cross_sections import (
    ScalarMediatorCrossSection,
)
from hazma.scalar_mediator._scalar_mediator_fsr import ScalarMediatorFSR
from hazma.scalar_mediator._scalar_mediator_positron_spectra import (
    ScalarMediatorPositronSpectra,
)
from hazma.scalar_mediator._scalar_mediator_spectra import ScalarMediatorSpectra
from hazma.scalar_mediator._scalar_mediator_widths import ScalarMediatorWidths

import numpy as np


# Note that Theory must be inherited from AFTER all the other mixin classes,
# since they furnish definitions of the abstract methods in Theory.
class ScalarMediator(
    ScalarMediatorConstraints,
    ScalarMediatorCrossSection,
    ScalarMediatorFSR,
    ScalarMediatorPositronSpectra,
    ScalarMediatorSpectra,
    ScalarMediatorWidths,
    Theory,
):
    r"""
    Create a scalar mediator model object.

    Creates an object for the scalar mediator model given UV couplings from
    common UV complete models of a real scalar extension of the SM. The UV
    complete models are:

        1) Scalar mediator coupling to a new heavy quark. When the heavy quark
           is integrated out of the theory, the scalar obtains an effective
           coupling to gluons, leading to a coupling to pions through a
           dialation current.

        2) Scalar mediator mixing with the standard model higgs. The scalar
           mediator obtains couplings to the massive standard model states
           which will be `sin(theta) m / v_h` where theta is the mixing angle
           between the higgs and the scalar, m is the mass of the massive state
           and v_h is the higgs vev.  The scalar mediator also gets an
           effective coupling to gluons when the top quark is integrated out.

    Parameters
    ----------
    mx : float
        Dark matter mass.
    ms : float
        Mass of the scalar mediator.
    gsxx : float
        Coupling constant for scalar interactions with dark matter.
    gsff : float
        Constant of proportionality for the scalar's couplings to Standard
        Model fermions.
    gsGG : float
        Coupling constant for scalar interactions with gluons.
    gsFF : float
        Coupling constant for scalar interactions with photons.
    lam : float
        Mass scale in scalar's interactions with photons and gluons.
    """

    def __init__(self, mx, ms, gsxx, gsff, gsGG, gsFF, lam):
        self._mx = mx
        self._ms = ms
        self._gsxx = gsxx
        self._gsff = gsff
        self._gsGG = gsGG
        self._gsFF = gsFF
        self._lam = lam
        self.width_s = 0.0
        self.vs = self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    @property
    def mx(self):
        return self._mx

    @mx.setter
    def mx(self, mx):
        self._mx = mx
        self.vs = self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    @property
    def ms(self):
        return self._ms

    @ms.setter
    def ms(self, ms):
        self._ms = ms
        self.vs = self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    @property
    def gsxx(self):
        return self._gsxx

    @gsxx.setter
    def gsxx(self, gsxx):
        self._gsxx = gsxx
        self.vs = self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    @property
    def gsff(self):
        return self._gsff

    @gsff.setter
    def gsff(self, gsff):
        self._gsff = gsff
        self.vs = self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    @property
    def gsGG(self):
        return self._gsGG

    @gsGG.setter
    def gsGG(self, gsGG):
        self._gsGG = gsGG
        self.vs = self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    @property
    def gsFF(self):
        return self._gsFF

    @gsFF.setter
    def gsFF(self, gsFF):
        self._gsFF = gsFF
        self.vs = self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    @property
    def lam(self):
        return self._lam

    @lam.setter
    def lam(self, lam):
        self._lam = lam
        self.vs = self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    def compute_vs(self):
        """
        Compute the value of the scalar mediator vacuum expectation value.

        Notes
        -----
        Warning! Currently this function simply returns zero. This is a very
        good approximation in areas of parameters space which are allowed by
        experimental observations.

        Returns
        -------
        vs: float
            vacuum expectation value of scalar mediator.
        """
        if 3 * self.gsff + 2 * self.gsGG == 0:
            vs = 0.0
        else:
            # trM = muq + mdq + msq

            # ms = self.ms
            # gsff = self.gsff
            # gsGG = self.gsGG
            # Lam = self.lam

            # self.vs = (-3 * ms * vh +
            #           np.sqrt(4 * b0 * fpi**2 *
            #                   (3 * gsff + 2 * gsGG)**2 * trM +
            #                   9 * ms**2 * vh**2)) / (6 * gsff * ms +
            #                                          4 * gsGG * ms)
            vs = 0.0

        return vs

    def compute_width_s(self):
        """Updates the scalar's total width.
        """
        self.width_s = self.partial_widths()["total"]

    def __fpiT(self, vs):
        """Returns the Lagrangian parameter __fpiT.
        """
        return fpi / np.sqrt(1.0 + 4.0 * self._gsGG * vs / (9.0 * vh))

    def __b0T(self, vs, fpiT):
        """Returns the Lagrangian parameter __b0T.
        """
        return (
            b0
            * (fpi / fpiT) ** 2
            / (1.0 + vs / vh * (2.0 * self._gsGG / 3.0 + self._gsff))
        )

    def __msT(self, fpiT, b0T):
        """Returns the Lagrangian parameter __msT.
        """
        trM = muq + mdq + msq

        return np.sqrt(
            self._ms ** 2
            - 16.0
            * self._gsGG
            * b0T
            * fpiT ** 2
            / (81.0 * vh ** 2)
            * (2.0 * self._gsGG - 9.0 * self._gsff)
            * trM
        )

    @staticmethod
    def list_annihilation_final_states():
        """
        Return a list of the available final states.

        Returns
        -------
        fs : array-like
            Array of the available final states.
        """
        return ["mu mu", "e e", "g g", "pi0 pi0", "pi pi", "s s"]


class HiggsPortal(ScalarMediator):
    r"""Create a ``ScalarMediator`` object with Higgs Portal couplings.

    The couplings in the full scalar model are defined by::

        gsff = sin(theta)
        gsGG = 3 sin(theta)
        gsFF = -5/6 sin(theta)
        Lam = vh

    where theta is the mixing angle between the Standard Model Higgs
    and the scalar mediator.

    Parameters
    ----------
    mx : float
        Mass of the dark matter.
    ms : float
        Mass of the scalar mediator.
    gsxx : float
        Coupling of scalar mediator to dark matter.
    stheta : float
        Sine of the mixing angle between the Standard Model Higgs
        and the scalar mediator.
    """

    def __init__(self, mx, ms, gsxx, stheta):
        self._lam = vh
        self._stheta = stheta

        super(HiggsPortal, self).__init__(
            mx, ms, gsxx, stheta, 3.0 * stheta, -5.0 * stheta / 6.0, vh
        )

    @property
    def stheta(self):
        return self._stheta

    @stheta.setter
    def stheta(self, stheta):
        self._stheta = stheta
        self._gsff = stheta
        self._gsGG = 3.0 * stheta
        self._gsFF = -5.0 * stheta / 6.0
        self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    # Hide underlying properties' setters
    @ScalarMediator.gsff.setter
    def gsff(self, gsff):
        raise AttributeError("Cannot set gsff")

    @ScalarMediator.gsGG.setter
    def gsGG(self, gsGG):
        raise AttributeError("Cannot set gsGG")

    @ScalarMediator.gsFF.setter
    def gsFF(self, gsFF):
        raise AttributeError("Cannot set gsFF")


class HeavyQuark(ScalarMediator):
    r"""
    Create a ScalarMediator object with heavy quark couplings.

    Creates an object for the scalar mediator model with the following
    specific coupling definitions:

        gsff = 0
        gsGG = gsQ
        gsFF = 0
        Lam = mQ

    where gsQ is the coupling of the heavy quark to the scalar mediator
    (-gsQ S Qbar Q) and mQ is the mass of the heavy quark.

    Parameters
    ----------
    mx : float
        Mass of the dark matter.
    ms : float
        Mass of the scalar mediator.
    gsxx : float
        Coupling of scalar mediator to dark matter.
    gsQ : float
        Coupling of the heavy quark to the scalar mediator.
    mQ : float
        Mass of the heavy quark.
    QQ : float
        Charge of the heavy quark.
    """

    def __init__(self, mx, ms, gsxx, gsQ, mQ, QQ):
        self._gsQ = gsQ
        self._mQ = mQ
        self._QQ = QQ

        super(HeavyQuark, self).__init__(
            mx, ms, gsxx, 0.0, gsQ, 2.0 * gsQ * QQ ** 2, mQ
        )

    @staticmethod
    def list_annihilation_final_states():
        return ["g g", "pi0 pi0", "pi pi", "s s"]

    @property
    def gsQ(self):
        return self._gsQ

    @gsQ.setter
    def gsQ(self, gsQ):
        self._gsQ = gsQ
        self._gsGG = gsQ
        self._gsFF = 2.0 * gsQ * self._QQ ** 2
        self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    @property
    def mQ(self):
        return self._mQ

    @mQ.setter
    def mQ(self, mQ):
        self._mQ = mQ
        self._lam = mQ
        self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    @property
    def QQ(self):
        return self._QQ

    @QQ.setter
    def QQ(self, QQ):
        self._QQ = QQ
        self._gsFF = 2.0 * self._gsQ * QQ ** 2
        self.compute_vs()
        self.compute_width_s()  # vs MUST be computed first

    # Hide underlying properties' setters
    @ScalarMediator.gsff.setter
    def gsff(self, gsff):
        raise AttributeError("Cannot set gsff")

    @ScalarMediator.gsGG.setter
    def gsGG(self, gsGG):
        raise AttributeError("Cannot set gsGG")

    @ScalarMediator.gsFF.setter
    def gsFF(self, gsFF):
        raise AttributeError("Cannot set gsFF")



import numpy as np
cimport numpy as np
from decay_muon cimport CSpectrum  as muspec

cdef double muon_spectrum(double)
cdef double gamma(double, double)
cdef double beta(double, double)
cdef double eng_gam_max(double)
cdef double integrand(double, double, double, str)

cdef double CSpectrumPoint(double, double, str)
cdef np.ndarray[np.float64_t, ndim=1] CSpectrum(np.ndarray, double, str)



from hazma.theory import Theory

from hazma.vector_mediator._vector_mediator_cross_sections import (
    VectorMediatorCrossSections,
)
from hazma.vector_mediator._vector_mediator_fsr import VectorMediatorFSR
from hazma.vector_mediator._vector_mediator_positron_spectra import (
    VectorMediatorPositronSpectra,
)
from hazma.vector_mediator._vector_mediator_spectra import VectorMediatorSpectra
from hazma.vector_mediator._vector_mediator_widths import VectorMediatorWidths
from ..parameters import qe, Qu, Qd, Qe


# Note that Theory must be inherited from AFTER all the other mixin classes,
# since they furnish definitions of the abstract methods in Theory.
class VectorMediator(
    VectorMediatorCrossSections,
    VectorMediatorFSR,
    VectorMediatorPositronSpectra,
    VectorMediatorSpectra,
    VectorMediatorWidths,
    Theory,
):
    r"""
    Create a VectorMediator object with generic couplings.

    Parameters
    ----------
    mx : float
        Mass of the dark matter.
    mv : float
        Mass of the vector mediator.
    gvxx : float
        Coupling of vector mediator to dark matter.
    gvuu : float
        Coupling of vector mediator to the up quark.
    gvdd : float
        Coupling of vector mediator to the down quark.
    gvss : float
        Coupling of vector mediator to the strange quark.
    gvee : float
        Coupling of vector mediator to the electron.
    gvmumu : float
        Coupling of vector mediator to the muon.
    """

    def __init__(self, mx, mv, gvxx, gvuu, gvdd, gvss, gvee, gvmumu):
        self._mx = mx
        self._mv = mv
        self._gvxx = gvxx
        self._gvuu = gvuu
        self._gvdd = gvdd
        self._gvss = gvss
        self._gvee = gvee
        self._gvmumu = gvmumu
        self.compute_width_v()

    @property
    def mx(self):
        return self._mx

    @mx.setter
    def mx(self, mx):
        self._mx = mx
        self.compute_width_v()

    @property
    def mv(self):
        return self._mv

    @mv.setter
    def mv(self, mv):
        self._mv = mv
        self.compute_width_v()

    @property
    def gvxx(self):
        return self._gvxx

    @gvxx.setter
    def gvxx(self, gvxx):
        self._gvxx = gvxx
        self.compute_width_v()

    @property
    def gvuu(self):
        return self._gvuu

    @gvuu.setter
    def gvuu(self, gvuu):
        self._gvuu = gvuu
        self.compute_width_v()

    @property
    def gvdd(self):
        return self._gvdd

    @gvdd.setter
    def gvdd(self, gvdd):
        self._gvdd = gvdd
        self.compute_width_v()

    @property
    def gvss(self):
        return self._gvss

    @gvss.setter
    def gvss(self, gvss):
        self._gvss = gvss
        self.compute_width_v()

    @property
    def gvee(self):
        return self._gvee

    @gvee.setter
    def gvee(self, gvee):
        self._gvee = gvee
        self.compute_width_v()

    @property
    def gvmumu(self):
        return self._gvmumu

    @gvmumu.setter
    def gvmumu(self, gvmumu):
        self._gvmumu = gvmumu
        self.compute_width_v()

    def compute_width_v(self):
        """Recomputes the scalar's total width."""
        self.width_v = self.partial_widths()["total"]

    @staticmethod
    def list_annihilation_final_states():
        """
        Return a list of the available final states.

        Returns
        -------
        fs : array-like
            Array of the available final states.
        """
        return ["mu mu", "e e", "pi pi", "pi0 g", "pi0 v", "v v"]

    def constraints(self):
        pass

    def constrain(self, p1, p1_vals, p2, p2_vals, ls_or_img="image"):
        pass


class KineticMixing(VectorMediator):
    r"""
    Create a ``VectorMediator`` object with kinetic mixing couplings.

    The couplings are defined as::

        gvuu = Qu qe eps
        gvdd = Qd qe eps
        gvss = Qd qe eps
        gvee = Qe qe eps
        gvmumu = Qe qe eps

    where Qu, Qd and Qe are the up-type quark, down-type quark and
    lepton electic charges in units of the electric charge, qe is the
    electric charge and eps is the kinetic mixing parameter.

    Parameters
    ----------
    mx : float
        Mass of the dark matter.
    mv : float
        Mass of the vector mediator.
    gvxx : float
        Coupling of vector mediator to dark matter.
    eps : float
        Kinetic mixing parameter.
    """

    def __init__(self, mx, mv, gvxx, eps):
        self._eps = eps

        super(KineticMixing, self).__init__(
            mx,
            mv,
            gvxx,
            -Qu * eps * qe,
            -Qd * eps * qe,
            -Qd * eps * qe,
            -Qe * eps * qe,
            -Qe * eps * qe,
        )

    @property
    def eps(self):
        return self._eps

    @eps.setter
    def eps(self, eps):
        self._eps = eps
        self._gvuu = -Qu * eps * qe
        self._gvdd = -Qd * eps * qe
        self._gvss = -Qd * eps * qe
        self._gvee = -Qe * eps * qe
        self._gvmumu = -Qe * eps * qe
        self.compute_width_v()

    # Hide underlying properties' setters
    @VectorMediator.gvuu.setter
    def gvuu(self, gvuu):
        raise AttributeError("Cannot set gvuu")

    @VectorMediator.gvdd.setter
    def gvdd(self, gvdd):
        raise AttributeError("Cannot set gvdd")

    @VectorMediator.gvss.setter
    def gvss(self, gvss):
        raise AttributeError("Cannot set gvss")

    @VectorMediator.gvee.setter
    def gvee(self, gvee):
        raise AttributeError("Cannot set gvee")

    @VectorMediator.gvmumu.setter
    def gvmumu(self, gvmumu):
        raise AttributeError("Cannot set gvmumu")


class QuarksOnly(VectorMediator):
    r"""
    Create a VectorMediator object with only quark couplings.

    Parameters
    ----------
    mx : float
        Mass of the dark matter.
    mv : float
        Mass of the vector mediator.
    gvxx : float
        Coupling of vector mediator to dark matter.
    gvuu : float
        Coupling of vector mediator to the up quark.
    gvdd : float
        Coupling of vector mediator to the down quark.
    gvss : float
        Coupling of vector mediator to the strange quark.
    """

    def __init__(self, mx, mv, gvxx, gvuu, gvdd, gvss):
        super(QuarksOnly, self).__init__(mx, mv, gvxx, gvuu, gvdd, gvss, 0.0, 0.0)

    @staticmethod
    def list_annihilation_final_states():
        return ["pi pi", "pi0 g", "pi0 v", "v v"]

    # Hide underlying properties' setters
    @VectorMediator.gvee.setter
    def gvee(self, gvee):
        raise AttributeError("Cannot set gvee")

    @VectorMediator.gvmumu.setter
    def gvmumu(self, gvmumu):
        raise AttributeError("Cannot set gvmumu")



from hazma.parameters import alpha_em, qe
from hazma.parameters import charged_pion_mass as mpi
from hazma.parameters import electron_mass as me
from hazma.parameters import muon_mass as mmu

from cmath import sqrt, log, pi
import numpy as np


class VectorMediatorFSR:
    def __dnde_xx_to_v_to_ffg(self, egam, Q, f):
        """Return the fsr spectra for fermions from decay of vector mediator.

        Computes the final state radiaton spectrum value dNdE from a vector
        mediator given a gamma ray energy of `egam`, center of mass energy `Q`
        and final state fermion `f`.

        Paramaters
        ----------
        egam : float
            Gamma ray energy.
        Q: float
            Center of mass energy of mass of off-shell vector mediator.
        f : float
            Name of the final state fermion: "e" or "mu".

        Returns
        -------
        spec_val : float
            Spectrum value dNdE from vector mediator.
        """
        if f == "e":
            mf = me
        elif f == "mu":
            mf = mmu

        e, m = egam / Q, mf / Q

        s = Q ** 2 - 2.0 * Q * egam

        ret_val = 0.0

        if 4.0 * mf ** 2 <= s <= Q ** 2 and Q > 2.0 * mf and Q > 2.0 * self.mx:
            ret_val = -(
                alpha_em
                * (
                    4.0
                    * sqrt(1.0 - 2.0 * e - 4.0 * m ** 2)
                    * (1.0 - 2.0 * m ** 2 + 2.0 * e * (-1 + e + 2.0 * m ** 2))
                    + sqrt(1.0 - 2.0 * e)
                    * (1.0 + 2.0 * (-1 + e) * e - 4.0 * e * m ** 2 - 4.0 * m ** 4)
                    * (
                        log(1.0 - 2.0 * e)
                        - 4.0
                        * log(sqrt(1.0 - 2.0 * e) + sqrt(1.0 - 2.0 * e - 4.0 * m ** 2))
                        + 2.0
                        * log(
                            (sqrt(1.0 - 2.0 * e) - sqrt(1.0 - 2.0 * e - 4.0 * m ** 2))
                            * (1.0 - sqrt(1.0 + (4.0 * m ** 2) / (-1 + 2.0 * e)))
                        )
                    )
                )
            ) / (
                2.0
                * e
                * (1.0 + 2.0 * m ** 2)
                * sqrt((-1 + 2.0 * e) * (-1 + 4.0 * m ** 2))
                * pi
                * Q
            )

            assert ret_val.imag == 0.0
            ret_val = ret_val.real

        assert ret_val >= 0

        return ret_val.real

    def dnde_xx_to_v_to_ffg(self, egam, Q, f):
        """Return the fsr spectra for fermions from decay of vector mediator.

        Computes the final state radiaton spectrum value dNdE from a vector
        mediator given a gamma ray energy of `egam`, center of mass energy `Q`
        and final state fermion `f`.

        Paramaters
        ----------
        egam : float
            Gamma ray energy.
        Q: float
            Center of mass energy of mass of off-shell vector mediator.
        f : float
            Mass of the final state fermion.

        Returns
        -------
        spec_val : float
            Spectrum value dNdE from vector mediator.
        """
        if hasattr(egam, "__len__"):
            return np.array([self.__dnde_xx_to_v_to_ffg(e, Q, f) for e in egam])
        else:
            return self.__dnde_xx_to_v_to_ffg(egam, Q, f)

    def __dnde_xx_to_v_to_pipig(self, egam, Q):
        """Unvectorized dnde_xx_to_v_to_pipig"""
        mx = self.mx

        mupi = mpi / Q
        x = 2.0 * egam / Q
        xmin = 0.0
        xmax = 1 - 4.0 * mupi ** 2

        if x < xmin or x > xmax or Q < 2 * mpi or Q < 2.0 * mx:
            return 0.0

        coeff = qe ** 2 / (4.0 * (1 - 4 * mupi ** 2) ** 1.5 * pi ** 2)

        dynamic = (
            (
                2
                * sqrt(1 - 4 * mupi ** 2 - x)
                * (-1 - 4 * mupi ** 2 * (-1 + x) + x + x ** 2)
            )
            / sqrt(1 - x)
            + (-1 + 4 * mupi ** 2)
            * (-1 + 2 * mupi ** 2 + x)
            * log(
                (1 + sqrt(1 - x) * sqrt(1 - 4 * mupi ** 2 - x) - x) ** 2
                / (-1 + sqrt(1 - x) * sqrt(1 - 4 * mupi ** 2 - x) + x) ** 2
            )
        ) / x

        ret_val = 2.0 * dynamic * coeff / Q

        assert ret_val.imag == 0.0
        assert ret_val.real >= 0
        return ret_val.real

    def dnde_xx_to_v_to_pipig(self, eng_gams, Q):
        """
        Returns the gamma ray energy spectrum for two fermions annihilating
        into two charged pions and a photon.

        Parameters
        ----------
        eng_gam : numpy.ndarray or double
            Gamma ray energy.
        Q : double
            Center of mass energy, or sqrt((ppip + ppim + pg)^2).

        Returns
        -------
        Returns gamma ray energy spectrum for
        :math:`\chi\bar{\chi}\to\pi^{+}\pi^{-}\gamma` evaluated at the gamma
        ray energy(ies).

        """

        if hasattr(eng_gams, "__len__"):
            return np.array(
                [self.__dnde_xx_to_v_to_pipig(eng_gam, Q) for eng_gam in eng_gams]
            )
        else:
            return self.__dnde_xx_to_v_to_pipig(eng_gams, Q)



import numpy as np

from hazma.positron_spectra import charged_pion as pspec_charged_pion
from hazma.positron_spectra import muon as pspec_muon
from hazma.scalar_mediator.scalar_mediator_positron_spec import dnde_decay_s


class ScalarMediatorPositronSpectra:
    def dnde_pos_pipi(self, positron_energies, cme):
        """
        Positron/electron spectrum from dark matter annihilating into charged
        pions.

        Parameters
        ----------
        positron_energies: float or np.array
            Energies of the positrons/electrons
        cme: float
            Center of mass energy

        Returns
        -------
        dnde: float or np.array
            Positron spectrum evaluated at the `positron_energies`.
        """
        return pspec_charged_pion(positron_energies, cme / 2.0)

    def dnde_pos_mumu(self, positron_energies, cme):
        """
        Positron/electron spectrum from dark matter annihilating into muons

        Parameters
        ----------
        positron_energies: float or np.array
            Energies of the positrons/electrons
        cme: float
            Center of mass energy

        Returns
        -------
        dnde: float or np.array
            Positron spectrum evaluated at the `positron_energies`.
        """
        return pspec_muon(positron_energies, cme / 2.0)

    def dnde_pos_ss(self, positron_energies, cme, fs="total"):
        """
        Positron/electron spectrum from dark matter annihilating into muons

        Parameters
        ----------
        positron_energies: float or np.array
            Energies of the positrons/electrons
        cme: float
            Center of mass energy
        fs: str {'total'}
            String for which final states to consider when computing scalar
            mediator decay spectrum. Options are 'total', 'pi pi' or 'mu mu'.

        Returns
        -------
        dnde: float or np.array
            Positron spectrum evaluated at the `positron_energies`.
        """
        scalar_energy = cme / 2.0
        pws = self.partial_widths()

        if pws["total"] != 0:
            pw_array = np.array([pws["e e"], pws["mu mu"], pws["pi pi"]])
            pw_array /= pws["total"]

            # Factor of 2 since S is self-conjugate
            return 2.0 * dnde_decay_s(
                positron_energies, scalar_energy, self.ms, pw_array, fs
            )
        else:
            return np.zeros_like(positron_energies)

    def positron_spectrum_funcs(self):
        return {
            "mu mu": self.dnde_pos_mumu,
            "pi pi": self.dnde_pos_pipi,
            "s s": self.dnde_pos_ss,
        }

    def positron_lines(self, cme):
        """
        Positron/electron lines from dark matter annihilating into
        electrons/positrons.

        Parameters
        ----------
        cme: float
            Center of mass energy

        Returns
        -------
        dnde: dict
            Dictionary of dictionaries. Each sub-dictionary contains the
            location of the line and the branching fraction for the
            corresponding dark matter annihilation process.
        """
        bf = self.annihilation_branching_fractions(cme)["e e"]

        return {"e e": {"energy": cme / 2.0, "bf": bf}}



class AxialVectorMediatorParameters:
    def __init__(self, mx, ma, gaxx, gaff):
        self._mx = mx
        self._ma = ma
        self._gaxx = gaxx
        self._gaff = gaff

    @property
    def mx(self):
        return self._mx

    @mx.setter
    def mx(self, mx):
        self._mx = mx

    @property
    def ma(self):
        return self._ma

    @ma.setter
    def ms(self, ma):
        self._ma = ma

    @property
    def gaxx(self):
        return self._gaxx

    @gaxx.setter
    def gaxx(self, gaxx):
        self._gaxx = gaxx

    @property
    def gaff(self):
        return self._gaff

    @gaff.setter
    def gaff(self, gaff):
        self._gaff = gaff



"""Module for computing gamma ray spectra from a many-particle final state.

@author - Logan Morrison and Adam Coogan
@date - December 2017
"""
import numpy as np
cimport numpy as np
import cython
import multiprocessing as mp

from hazma import rambo
from hazma.rambo import compute_annihilation_cross_section
from hazma.rambo import compute_decay_width

from hazma.decay import muon as dm
from hazma.decay import electron as de

from hazma.decay import neutral_pion as dnp
from hazma.decay import charged_pion as dcp

from hazma.decay import charged_kaon as dck
from hazma.decay import long_kaon as dlk
from hazma.decay import short_kaon as dsk

include "../decay_helper_functions/parameters.pxd"

cdef spec_dict = {'muon': dm, 'electron': de, 'neutral_pion': dnp,
                  'charged_pion': dcp, 'charged_kaon': dck,
                  'long_kaon': dlk, 'short_kaon': dsk}

cdef mass_dict = {'muon': MASS_MU, 'electron': MASS_E,
                  'neutral_pion': MASS_PI0, 'charged_pion': MASS_PI, 'charged_kaon': MASS_K, 'long_kaon': MASS_K0,
                  'short_kaon': MASS_K0}

cdef dict cspec_dict = spec_dict
cdef dict cmass_dict = mass_dict

cdef np.ndarray names_to_masses(np.ndarray names):
    """Returns the masses of particles given a list of their names.

    Parameters
    ----------
    names : np.ndarray
        List of names of the final state particles.

    Returns
    -------
    masses : np.ndarray
        List of masses of the final state particles. For
        example, ['electron', 'muon'] -> [0.510998928, 105.6583715].
    """
    cdef int i
    cdef int size = len(names)
    cdef np.ndarray masses = np.zeros(size, dtype=np.float64)

    for i in range(size):
        masses[i] = cmass_dict[names[i]]
    return masses

def __gen_spec(name, eng, eng_gams, norm, verbose=False):
    """
    c-function used by ``gamma`` and ``gamma_point`` to generate spectrum
    values.
    """
    if verbose is True:
        print("creating {} spectrum with energy {}".format(name, eng))
    return norm * cspec_dict[name](eng_gams, eng)

def __gen_spec_2body(particles, cme, eng_gams):
    masses = names_to_masses(particles)

    m1 = masses[0]
    m2 = masses[1]

    E1 = (cme**2 + m1**2 - m2**2) / (2 * cme)
    E2 = (cme**2 - m1**2 + m2**2) / (2 * cme)

    spec = cspec_dict[particles[0]](eng_gams, E1)
    spec += cspec_dict[particles[1]](eng_gams, E2)

    return spec

@cython.boundscheck(False)
@cython.wraparound(False)
def gamma(np.ndarray particles, double cme,
          np.ndarray eng_gams, mat_elem_sqrd=lambda k_list: 1.0,
          int num_ps_pts=10000, int num_bins=25, verbose=False):
    """Returns total gamma ray spectrum from final state particles.

    Parameters
    ----------
    particles : np.ndarray[string, ndim=1]
        1-D array of strings containing the final state particle names. The
        accepted particle names are: "muon", "electron", "neutral_pion",
        "charged_pion", "long_kaon", "short_kaon" and "charged_kaon".
    cme : double
        Center of mass energy of the final state.
    eng_gams : np.ndarray[double, ndim=1]
        List of gamma ray energies to compute spectra at.
    mat_elem_sqrd : double(*)(np.ndarray)
        Function for the matrix element squared of the proccess. Must be
        a function taking in a list of four momenta of size (num_fsp, 4).
        Default value is a flat matrix element.
    num_ps_pts : int
        Number of phase space points to use.
    num_bins : int
        Number of bins to use.

    Returns
    -------
    spec : np.ndarray[double, ndim=1]
        1-D array of total gamma ray spectrum from all final state particles.
    """

    if len(particles) == 1:
        return cspec_dict[particles[0]](eng_gams, cme)

    if len(particles) == 2:
        return __gen_spec_2body(particles, cme, eng_gams)

    cdef int i, j
    cdef int num_fsp
    cdef int num_engs
    cdef np.ndarray masses
    cdef np.ndarray hist
    cdef double prefactor

    masses = names_to_masses(particles)

    num_fsp = len(masses)
    num_engs = len(eng_gams)

    hist = rambo.generate_energy_histogram(masses, cme, num_ps_pts, mat_elem_sqrd, num_bins, density=True)[0]

    cpdef int num_cpus = int(np.floor(mp.cpu_count() * 0.75))

    p = mp.Pool(num_cpus)
    specs = []

    for i in range(num_bins):
        for j in range(num_fsp):
            part = particles[j]  # particle name
            part_eng = hist[j, 0, i]  # particle energy

            # Normalize spectrum: Need to multiply by the probability of
            # particle having energy part_eng. Since we are essentially
            # integrating over the energy probability distribution, we need to
            # multiply by (b - a) / N, where a = e_min, b = e_max and
            # N = num_bins.
            norm = (hist[j, 0, -1] - hist[j, 0, 0]) / num_bins * hist[j, 1, i]

            specs.append(p.apply_async(__gen_spec, \
                                       (part, part_eng, eng_gams, norm, verbose)))

    p.close()
    p.join()

    return sum([spec.get() for spec in specs])

@cython.boundscheck(False)
@cython.wraparound(False)
def gamma_point(np.ndarray particles, double cme,
                double eng_gam, mat_elem_sqrd=lambda k_list: 1.0,
                int num_ps_pts=1000, int num_bins=25):
    """Returns total gamma ray spectrum from final state particles.

    Parameters
    ----------
    particles : np.ndarray[string, ndim=1]
        1-D array of strings containing the final state particle names. The
        accepted particle names are: "muon", "electron", "neutral_pion",
        "charged_pion", "long_kaon", "short_kaon" and "charged_kaon".
    cme : double
        Center of mass energy of the final state.
    eng_gam : double
        Gamma ray energy to evaluate spectrum at.
    mat_elem_sqrd : double(*)(np.ndarray)
        Function for the matrix element squared of the proccess. Must be
        a function taking in a list of four momenta of size (num_fsp, 4).
        Default value is a flat matrix element.
    num_ps_pts : int
        Number of phase space points to use.
    num_bins : int
        Number of bins to use.

    Returns
    -------
    spec : double
        Spectrum evaluated at ``eng_gam``.
    """
    cdef int i, j
    cdef int num_fsp
    cdef np.ndarray masses
    cdef np.ndarray hist
    cdef double spec_val = 0.0
    cdef double prefactor

    masses = names_to_masses(particles)

    num_fsp = len(masses)

    hist = rambo.generate_energy_histogram(masses, cme, num_ps_pts, mat_elem_sqrd, num_bins)[0]

    for i in range(num_bins):
        for j in range(num_fsp):
            if particles[j] is 'electron':
                spec_val += 0.0
            if particles[j] is 'muon':
                spec_val += hist[j, 1, i] * \
                            dm(eng_gam, hist[j, 0, i]) * \
                            (hist[j, 0, -1] - hist[j, 0, 0]) / num_bins
            if particles[j] is 'charged_pion':
                spec_val += hist[j, 1, i] * \
                            dcp(eng_gam, hist[j, 0, i], "total") * \
                            (hist[j, 0, -1] - hist[j, 0, 0]) / num_bins
            if particles[j] is 'neutral_pion':
                spec_val += hist[j, 1, i] * \
                            dnp(eng_gam, hist[j, 0, i]) * \
                            (hist[j, 0, -1] - hist[j, 0, 0]) / num_bins
            if particles[j] is 'charged_kaon':
                spec_val += hist[j, 1, i] * \
                            dck(eng_gam, hist[j, 0, i], "total") * \
                            (hist[j, 0, -1] - hist[j, 0, 0]) / num_bins
            if particles[j] is 'short_kaon':
                spec_val += hist[j, 1, i] * \
                            dsk(eng_gam, hist[j, 0, i], "total") * \
                            (hist[j, 0, -1] - hist[j, 0, 0]) / num_bins
            if particles[j] is 'long_kaon':
                spec_val += hist[j, 1, i] * \
                            dlk(eng_gam, hist[j, 0, i], "total") * \
                            (hist[j, 0, -1] - hist[j, 0, 0]) / num_bins

    return spec_val * prefactor



import numpy as np

from hazma.positron_spectra import muon as pspec_muon
from hazma.positron_spectra import positron_decay

from hazma.parameters import charged_pion_mass as mpi
from hazma.parameters import neutral_pion_mass as mpi0


class PseudoScalarMediatorPositronSpectra:
    def dnde_pos_pi0pipi(self, eng_ps, cme, num_ps_pts=1000):
        if cme <= 2.0 * mpi + mpi0:
            return np.zeros(len(eng_ps), dtype=float)

        return positron_decay(
            ["charged_pion", "charged_pion", "neutral_pion"],
            cme,
            eng_ps,
            num_ps_pts=1000,
            mat_elem_sqrd=self.msqrd_xx_to_p_to_pm0,
        )

    def dnde_pos_mumu(eng_ps, cme):
        return pspec_muon(eng_ps, cme / 2.0)

    def positron_spectra(self, eng_ps, cme):
        """Computes continuum part of positron spectrum from DM annihilation.

        Parameters
        ----------
        eng_ps : array-like
            Positron energies at which to compute the spectrum.
        cme : float
            Center of mass energy.

        Returns
        -------
        specs : dict
            Dictionary of positron spectra. The keys are the final states
            producing contributions to the continuum positron spectrum and
            'total'.
        """
        # Compute branching fractions
        bfs = self.annihilation_branching_fractions(cme)

        # Only compute the spectrum if the channel's branching fraction is
        # nonzero
        def spec_helper(bf, specfn):
            if bf != 0:
                return bf * specfn(eng_ps, cme)
            else:
                return np.zeros(eng_ps.shape)

        mumu_spec = spec_helper(bfs["mu mu"], self.dnde_pos_mumu)
        pi0pipi_spec = spec_helper(bfs["pi0 pi pi"], self.dnde_pos_pi0pipi)

        total = mumu_spec + pi0pipi_spec

        return {"total": total, "mu mu": mumu_spec, "pi0 pi pi": pi0pipi_spec}

    def positron_lines(self, cme):
        bf = self.annihilation_branching_fractions(cme)["e e"]

        return {"e e": {"energy": cme / 2.0, "bf": bf}}



from cmath import sqrt, pi

from hazma.parameters import (
    charged_pion_mass as mpi,
    neutral_pion_mass as mpi0,
    electron_mass as me,
    muon_mass as mmu,
    qe,
    fpi,
)


class VectorMediatorWidths:
    def width_v_to_pipi(self):
        mv = self.mv

        if mv > 2.0 * mpi:
            gvuu = self.gvuu
            gvdd = self.gvdd
            ret_val = ((gvdd - gvuu) ** 2 * (-4.0 * mpi ** 2 + mv ** 2) ** 1.5) / (
                16.0 * mv ** 2 * pi
            )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def width_v_to_pi0g(self):
        mv = self.mv

        if mv > mpi0:
            gvuu = self.gvuu
            gvdd = self.gvdd

            ret_val = (
                3.0
                * (
                    (gvdd + 2.0 * gvuu) ** 2
                    * (-mpi0 ** 2 + mv ** 2)
                    * (
                        -mpi0 ** 6
                        + 4.0 * mpi ** 4 * mv ** 2
                        + 12.0 * mpi0 ** 4 * mv ** 2
                        - 21.0 * mpi0 ** 2 * mv ** 4
                        + 4.0 * mv ** 6
                        + mpi ** 2
                        * (mpi0 ** 4 - 12.0 * mpi0 ** 2 * mv ** 2 + 13.0 * mv ** 4)
                    )
                    * qe ** 2
                )
                / (18432.0 * fpi ** 2 * mv ** 5 * pi ** 5)
            )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def width_v_to_xx(self):
        mv = self.mv
        mx = self.mx

        if mv > 2.0 * mx:
            gvxx = self.gvxx

            ret_val = (
                gvxx ** 2
                * sqrt(mv ** 2 - 4.0 * mx ** 2).real
                * (mv ** 2 + 2.0 * mx ** 2)
            ) / (4.0 * mv ** 2 * pi)

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def width_v_to_ff(self, f):
        if f == "e":
            mf = me
            gvll = self.gvee
        elif f == "mu":
            mf = mmu
            gvll = self.gvmumu
        else:
            return 0.0

        mv = self.mv

        if mv > 2.0 * mf:
            ret_val = (
                gvll ** 2
                * sqrt(mv ** 2 - 4.0 * mf ** 2).real
                * (2.0 * mf ** 2 + mv ** 2)
            ) / (4.0 * mv ** 2 * pi)

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def partial_widths(self):
        w_pipi = self.width_v_to_pipi()
        w_pi0g = self.width_v_to_pi0g()
        w_xx = self.width_v_to_xx()
        w_ee = self.width_v_to_ff("e")
        w_mumu = self.width_v_to_ff("mu")

        total = w_pipi + w_pi0g + w_xx + w_ee + w_mumu

        return {
            "pi pi": w_pipi,
            "pi0 g": w_pi0g,
            "x x": w_xx,
            "e e": w_ee,
            "mu mu": w_mumu,
            "total": total,
        }



"""
Generate histograms of the energies of particles.

* Author - Logan A. Morrison and Adam Coogan
* Date - December 2017
"""

import numpy as np
cimport numpy as np
import cython


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def space_to_energy_hist(np.ndarray[np.float64_t, ndim=2] pts, int num_ps_pts,
                         int num_fsp, int num_bins, density=False):
    """
    """
    cdef int i, j
    cdef int weight_index = 4 * num_fsp

    # Define arrays for various information.
    cdef np.ndarray[np.float64_t, ndim=1] weights
    cdef np.ndarray[np.float64_t, ndim=2] bins
    cdef np.ndarray[np.float64_t, ndim=2] engs
    cdef np.ndarray[np.float64_t, ndim=3] probs

    cdef np.ndarray[np.float64_t, ndim=2] means
    cdef np.ndarray[np.float64_t, ndim=2] vars
    cdef np.ndarray[np.float64_t, ndim=2] errs

    # Allocate arrays to store various information.
    weights = np.zeros(num_ps_pts, dtype=np.float64)
    bins = np.zeros((num_fsp, num_bins + 1), dtype=np.float64)
    engs = np.zeros((num_fsp, num_bins), dtype=np.float64)
    probs = np.zeros((num_fsp, 2, num_bins), dtype=np.float64)

    means = np.zeros((num_fsp, num_bins), dtype=np.float64)
    vars = np.zeros((num_fsp, num_bins), dtype=np.float64)
    errs = np.zeros((num_fsp, num_bins), dtype=np.float64)

    # Extract the weights from phase space points.
    for i in range(num_ps_pts):
        weights[i] = pts[i, weight_index]

    for i in range(num_fsp):

        means[i, :], bins[i, :] = np.histogram(pts[:, 4 * i], bins=num_bins,
                                               weights=weights, density=density)

        vars[i, :] = np.histogram(pts[:, 4 * i], bins=num_bins,
                                  weights=weights * weights, density=density)[0]

        errs[i, :] = np.sqrt((vars[i, :] - means[i, :]**2))

        # Normalize the distributions by the number of phase space points,
        # but only if density is False, otherwise the density is True case
        # will not be normalized to 1.
        if density is False:
            means[i, :] = means[i, :] * (1.0 / num_ps_pts)
            vars[i, :] = vars[i, :] * (1.0 / num_ps_pts)
            errs[i, :] = errs[i, :] * (1.0 / num_ps_pts)

    for i in range(num_fsp):
        for j in range(num_bins):
            engs[i, j] = (bins[i, j+1] + bins[i, j]) / 2

    for i in range(num_fsp):
        for j in range(num_bins):
            probs[i, 0, j] = engs[i, j]
            if density is True:
                probs[i, 1, j] = means[i, j]
            else:
                probs[i, 1, j] = means[i, j] / (bins[i, j+1] - bins[i, j])

    return probs, errs



import numpy as np

from hazma.positron_spectra import charged_pion as pspec_charged_pion
from hazma.positron_spectra import muon as pspec_muon
from hazma.vector_mediator.vector_mediator_positron_spec import dnde_decay_v


class VectorMediatorPositronSpectra:
    def dnde_pos_pipi(self, e_ps, e_cm):
        return pspec_charged_pion(e_ps, e_cm / 2.0)

    def dnde_pos_mumu(self, e_ps, e_cm):
        return pspec_muon(e_ps, e_cm / 2.0)

    # positron decay spectrum for chi chibar -> v v
    def dnde_pos_vv(self, e_ps, e_cm, fs="total"):
        # Each scalar gets half the COM energy
        e_v = e_cm / 2.0
        pws = self.partial_widths()

        if pws["total"] != 0:
            # dnde_decay_v relies on this ordering of the partial widths
            pw_array = np.array([pws["e e"], pws["mu mu"], pws["pi pi"]])
            pw_array /= pws["total"]

            # Factor of 2 since there are two V's
            return 2.0 * dnde_decay_v(e_ps, e_v, self.mv, pw_array, fs)
        else:
            return np.zeros_like(e_ps)

    def positron_spectrum_funcs(self):
        return {
            "mu mu": self.dnde_pos_mumu,
            "pi pi": self.dnde_pos_pipi,
            "v v": self.dnde_pos_vv,
        }

    def positron_lines(self, e_cm):
        bf = self.annihilation_branching_fractions(e_cm)["e e"]

        return {"e e": {"energy": e_cm / 2.0, "bf": bf}}



import numpy as np

from hazma.decay import muon
from hazma.decay import neutral_pion, charged_pion

from hazma.parameters import muon_mass as mmu
from hazma.parameters import electron_mass as me

from hazma.scalar_mediator.scalar_mediator_decay_spectrum import (
    dnde_decay_s,
    dnde_decay_s_pt,
)


class ScalarMediatorSpectra:
    def dnde_ee(self, e_gams, e_cm, spectrum_type="all"):
        if spectrum_type == "all":
            return self.dnde_ee(e_gams, e_cm, "fsr") + self.dnde_ee(
                e_gams, e_cm, "decay"
            )
        elif spectrum_type == "fsr":
            return self.dnde_xx_to_s_to_ffg(e_gams, e_cm, me)
        elif spectrum_type == "decay":
            return np.array([0.0 for _ in range(len(e_gams))])
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def dnde_mumu(self, e_gams, e_cm, spectrum_type="all"):
        if spectrum_type == "all":
            return self.dnde_mumu(e_gams, e_cm, "fsr") + self.dnde_mumu(
                e_gams, e_cm, "decay"
            )
        elif spectrum_type == "fsr":
            return self.dnde_xx_to_s_to_ffg(e_gams, e_cm, mmu)
        elif spectrum_type == "decay":
            return 2.0 * muon(e_gams, e_cm / 2.0)
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def dnde_pi0pi0(self, e_gams, e_cm, spectrum_type="all"):
        if spectrum_type == "all":
            return self.dnde_pi0pi0(e_gams, e_cm, "fsr") + self.dnde_pi0pi0(
                e_gams, e_cm, "decay"
            )
        if spectrum_type == "fsr":
            return np.array([0.0 for _ in range(len(e_gams))])
        if spectrum_type == "decay":
            return 2.0 * neutral_pion(e_gams, e_cm / 2.0)
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def dnde_pipi(self, e_gams, e_cm, spectrum_type="all"):
        if spectrum_type == "all":
            return self.dnde_pipi(e_gams, e_cm, "fsr") + self.dnde_pipi(
                e_gams, e_cm, "decay"
            )
        elif spectrum_type == "fsr":
            return self.dnde_xx_to_s_to_pipig(e_gams, e_cm)
        elif spectrum_type == "decay":
            return 2.0 * charged_pion(e_gams, e_cm / 2.0)
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def dnde_ss(self, e_gams, e_cm, fs="total"):
        # Each scalar gets half the COM energy
        e_s = e_cm / 2.0

        ms = self.ms
        pws = self.partial_widths()
        if pws["total"] != 0:
            pw_array = np.array(
                [pws["e e"], pws["mu mu"], pws["pi0 pi0"], pws["pi pi"], pws["g g"]],
                dtype=float,
            )
            pw_array /= pws["total"]

            # Factor of 2 since S is self-conjugate
            return 2.0 * dnde_decay_s(e_gams, e_s, ms, pw_array, fs)
        else:
            return np.zeros_like(e_gams)

    def spectrum_funcs(self):
        """
        Returns a dictionary of all the avaiable spectrum functions for
        a pair of initial state fermions with mass `mx` annihilating into
        each available final state.

        Each argument of the spectrum functions in `eng_gams`, an array
        of the gamma ray energies to evaluate the spectra at and `e_cm`, the
        center of mass energy of the process.
        """
        return {
            "mu mu": self.dnde_mumu,
            "e e": self.dnde_ee,
            "pi0 pi0": self.dnde_pi0pi0,
            "pi pi": self.dnde_pipi,
            "s s": self.dnde_ss,
        }

    def gamma_ray_lines(self, e_cm):
        bf = self.annihilation_branching_fractions(e_cm)["g g"]

        return {"g g": {"energy": e_cm / 2.0, "bf": bf}}



"""Module for computing fsr spectrum from a pseudo-scalar mediator.

@author - Logan Morrison and Adam Coogan.
@data - December 2017

"""
import numpy as np

from cmath import sqrt, log, pi

from hazma.parameters import alpha_em


class PseudoScalarMediatorFSR:
    def __dnde_xx_to_p_to_ffg(self, egam, Q, mf):
        """
        Returns the fsr spectra for fermions from decay of pseudo-scalar
        mediator.

        Computes the final state radiaton spectrum value dNdE from a
        pseudo-scalar mediator given a gamma ray energy of `eng_gam`,
        center of mass
        energy `cme` and final state fermion mass `mass_f`.

        Paramaters
        ----------
        eng_gam : float
            Gamma ray energy.
        cme: float
            Center of mass energy of mass of off-shell pseudo-scalar mediator.
        mass_f : float
            Mass of the final state fermion.

        Returns
        -------
        spec_val : float
            Spectrum value dNdE from pseudo-scalar mediator.
        """
        e, m = egam / Q, mf / Q

        s = Q ** 2 - 2.0 * Q * egam

        ret_val = 0.0

        if 4.0 * mf ** 2 <= s <= Q ** 2:
            ret_val = (
                2.0
                * alpha_em
                * (
                    -sqrt((-1.0 + 2.0 * e) * (-1.0 + 2.0 * e + 4.0 * m ** 2))
                    + (2.0 + 4.0 * (-1.0 + e) * e) * log(m)
                    + m ** 2
                    * (
                        2.0
                        * log(sqrt(1.0 - 2.0 * e) - sqrt(1.0 - 2.0 * e - 4.0 * m ** 2))
                        - log(
                            2.0
                            * (
                                1.0
                                - 2.0 * e
                                - 2.0 * m ** 2
                                + sqrt(
                                    (-1.0 + 2.0 * e) * (-1.0 + 2.0 * e + 4.0 * m ** 2)
                                )
                            )
                        )
                    )
                    + (1.0 + 2.0 * (-1.0 + e) * e)
                    * log(
                        -2.0
                        / (
                            -1.0
                            + 2.0 * e
                            + 2.0 * m ** 2
                            + sqrt((-1.0 + 2.0 * e) * (-1.0 + 2.0 * e + 4.0 * m ** 2))
                        )
                    )
                )
            ) / (e * sqrt(1.0 - 4.0 * m ** 2) * pi * Q)

            assert ret_val.imag == 0.0

            ret_val = ret_val.real

            assert ret_val >= 0.0

        return ret_val

    def dnde_xx_to_p_to_ffg(self, egam, Q, mf):
        """Returns the fsr spectra for fermions from decay of pseudo-scalar
        mediator.

        Computes the final state radiaton spectrum value dNdE from a
        pseudo-scalar mediator given a gamma ray energy of `eng_gam`,
        center of mass energy `cme` and final state fermion mass `mass_f`.

        Paramaters
        ----------
        eng_gam : float
            Gamma ray energy.
        cme: float
            Center of mass energy of mass of off-shell pseudo-scalar mediator.
        mass_f : float
            Mass of the final state fermion.

        Returns
        -------
        spec_val : float
            Spectrum value dNdE from pseudo-scalar mediator.
        """
        if hasattr(egam, "__len__"):
            return np.array([self.__dnde_xx_to_p_to_ffg(e, Q, mf) for e in egam])
        else:
            return self.__dnde_xx_to_p_to_ffg(egam, Q, mf)



from libc.math cimport sqrt
import numpy as np
cimport numpy as np
import cython

include "parameters.pxd"

"""
Module for computing the decay spectrum from neutral pion.
"""

@cython.cdivision(True)
cdef double CSpectrumPoint(double eng_gam, double eng_pi):
    """
    Returns decay spectrum for pi0 -> g g.
    """
    if eng_pi < MASS_PI0:
        return 0.0

    cdef float beta = sqrt(1.0 - (MASS_PI0 / eng_pi)**2)
    cdef float ret_val = 0.0

    if eng_pi * (1 - beta) / 2.0 <= eng_gam <= eng_pi * (1 + beta) / 2.0:
        ret_val = BR_PI0_TO_GG * 2.0 / (eng_pi * beta)

    return ret_val

@cython.cdivision(True)
cdef np.ndarray CSpectrum(np.ndarray eng_gam, double eng_pi):
    """
    Returns decay spectrum for pi0 -> g g.
    """

    cdef int num_pts = len(eng_gam)
    cdef np.ndarray spec = np.zeros(num_pts, dtype=np.float64)
    cdef int i

    for i in range(num_pts):
        spec[i] = CSpectrumPoint(eng_gam[i], eng_pi)

    return spec

@cython.cdivision(True)
def SpectrumPoint(double eng_gam, double eng_pi):
    """
    Returns decay spectrum for pi0 -> g g.
    """
    return CSpectrumPoint(eng_gam, eng_pi)

@cython.cdivision(True)
def Spectrum(np.ndarray eng_gam, double eng_pi):
    """
    Returns decay spectrum for pi0 -> g g.
    """
    return CSpectrum(eng_gam, eng_pi)



cimport decay_muon
cimport decay_charged_pion
cimport decay_neutral_pion
import numpy as np
cimport numpy as np
from scipy.integrate import quad
from libc.math cimport sqrt
import cython
import os
import sys
from .get_path import get_dir_path
include "parameters.pxd"
import warnings

"""
Module for computing the photon spectrum from radiative kaon decay.

Description:
    The charged kaon has many decay modes:

        k -> mu  + nu
        k -> pi  + pi0
        k -> pi  + pi  + pi
        k -> pi0 + e   + nu
        k -> pi0 + mu  + nu
        k -> pi  + pi0 + pi0

    For the the two-body final states, the sum of the decay spectra are
    computed given the known energies of the final state particles in the
    kaon's rest frame. The spectrum is then boosted into the lab frame.

    For the three-body final state, the energies of the final state
    particles are computed using RAMBO. The spectra for each final state
    particle is computed are each point in phases space in the charged kaon's rest frame and then spectra are summed over. The spectra is then boosted into the lab frame.
"""


data_path_total = os.path.join(get_dir_path(),
                               "interpolation_data",
                               "ckaon",
                               "charged_kaon_interp_total.dat")
data_path_0enu = os.path.join(get_dir_path(),
                              "interpolation_data",
                              "ckaon",
                              "charged_kaon_interp_0enu.dat")
data_path_0munu = os.path.join(get_dir_path(),
                               "interpolation_data",
                               "ckaon",
                               "charged_kaon_interp_0munu.dat")
data_path_00p = os.path.join(get_dir_path(),
                             "interpolation_data",
                             "ckaon",
                             "charged_kaon_interp_00p.dat")
data_path_mmug = os.path.join(get_dir_path(),
                              "interpolation_data",
                              "ckaon",
                              "charged_kaon_interp_mmug.dat")
data_path_munu = os.path.join(get_dir_path(),
                              "interpolation_data",
                              "ckaon",
                              "charged_kaon_interp_munu.dat")
data_path_p0 = os.path.join(get_dir_path(),
                            "interpolation_data",
                            "ckaon",
                            "charged_kaon_interp_p0.dat")
data_path_p0g = os.path.join(get_dir_path(),
                             "interpolation_data",
                             "ckaon",
                             "charged_kaon_interp_p0g.dat")
data_path_ppm = os.path.join(get_dir_path(),
                             "interpolation_data",
                             "ckaon",
                             "charged_kaon_interp_ppm.dat")

__e_gams_total, __spec_total = np.loadtxt(data_path_total, delimiter=',').T
__e_gams_0enu, __spec_0enu = np.loadtxt(data_path_0enu, delimiter=',').T
__e_gams_0munu, __spec_0munu = np.loadtxt(data_path_0munu, delimiter=',').T
__e_gams_00p, __spec_00p = np.loadtxt(data_path_00p, delimiter=',').T
__e_gams_mmug, __spec_mmug = np.loadtxt(data_path_mmug, delimiter=',').T
__e_gams_munu, __spec_munu = np.loadtxt(data_path_munu, delimiter=',').T
__e_gams_p0, __spec_p0 = np.loadtxt(data_path_p0, delimiter=',').T
__e_gams_p0g, __spec_p0g = np.loadtxt(data_path_p0g, delimiter=',').T
__e_gams_ppm, __spec_ppm = np.loadtxt(data_path_ppm, delimiter=',').T

cdef double __interp_spec(double eng_gam, str mode):
    """
    Intepolation function for the charged kaon data.

    Parameters
    ----------
    eng_gam : double
        Energy of the photon.
    mode : str {"total"}
        Optional. The decay mode to use.
    """
    if mode == "total":
        return np.interp(eng_gam, __e_gams_total, __spec_total)
    if mode == "0enu":
        return np.interp(eng_gam, __e_gams_0enu, __spec_0enu)
    if mode == "0munu":
        return np.interp(eng_gam, __e_gams_0munu, __spec_0munu)
    if mode == "00p":
        return np.interp(eng_gam, __e_gams_00p, __spec_00p)
    if mode == "mmug":
        return np.interp(eng_gam, __e_gams_mmug, __spec_mmug)
    if mode == "munu":
        return np.interp(eng_gam, __e_gams_munu, __spec_munu)
    if mode == "p0":
        return np.interp(eng_gam, __e_gams_p0, __spec_p0)
    if mode == "p0g":
        return np.interp(eng_gam, __e_gams_p0g, __spec_p0g)
    if mode == "ppm":
        return np.interp(eng_gam, __e_gams_ppm, __spec_ppm)



@cython.cdivision(True)
cdef double __integrand(double cl, double eng_gam, double eng_k, str mode):
    """
    Integrand for K -> X, where X is a any final state. The X's
    used are
        mu + nu
        pi  + pi0
        pi + pi + pi
        pi0 + mu + nu.
    When the ChargedKaon object is instatiated, the energies of the FSP are
    computed using RAMBO and energy distributions are formed. All the
    energies from the energy distributions are summed over against their
    weights.

    Keyword arguments::
        cl: Angle of photon w.r.t. charged kaon in lab frame.
        eng_gam: Energy of photon in laboratory frame.
        eng_k: Energy of kaon in laboratory frame.
    """
    cdef double gamma_k = eng_k / MASS_K
    cdef double beta_k = sqrt(1.0 - (MASS_K / eng_k)**2)
    cdef double eng_gam_k_rf = eng_gam * gamma_k * (1.0 - beta_k * cl)

    cdef double pre_factor \
        = 1.0 / (2.0 * gamma_k * (1.0 - beta_k * cl))

    return  __interp_spec(eng_gam_k_rf, mode)


cdef double CSpectrumPoint(double eng_gam, double eng_k, str mode):
    """
    Returns the radiative spectrum value from charged kaon at
    a single gamma ray energy.

    Keyword arguments::
        eng_gam: Energy of photon is laboratory frame.
        eng_k: Energy of charged kaon in laboratory frame.
    """
    message = 'Energy of kaon cannot be less than the kaon mass. Returning 0.'
    if eng_k < MASS_K:
        # raise warnings.warn(message, RuntimeWarning)
        return 0.0

    cdef double result = 0.0

    result = quad(__integrand, -1.0, 1.0, points=[-1.0, 1.0], \
                  args=(eng_gam, eng_k, mode), epsabs=10**-10., \
                  epsrel=10**-4.)[0]

    return result

@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.ndarray CSpectrum(np.ndarray[np.float64_t, ndim=1] eng_gams,
                          double eng_k, str mode):
    """
    Returns the radiative spectrum dNde from charged kaon for a
    list of gamma ray energies.

    Keyword arguments::
        eng_gams: List of energies of photon in laboratory frame.
        eng_k: Energy of charged kaon in laboratory frame.
    """

    cdef int numpts = len(eng_gams)

    cdef np.ndarray spec = np.zeros(numpts, dtype=np.float64)

    cdef int i = 0

    for i in range(numpts):
        spec[i] = CSpectrumPoint(eng_gams[i], eng_k, mode)

    return spec

def SpectrumPoint(double eng_gam, double eng_k, str mode):
    """
    Returns the radiative spectrum value from charged kaon at
    a single gamma ray energy.

    Keyword arguments::
        eng_gam: Energy of photon is laboratory frame.
        eng_k: Energy of charged kaon in laboratory frame.
    """

    return CSpectrumPoint(eng_gam, eng_k, mode)

@cython.boundscheck(False)
@cython.wraparound(False)
def Spectrum(np.ndarray[np.float64_t, ndim=1] eng_gams, double eng_k,
             str mode):
    """
    Returns the radiative spectrum dNde from charged kaon for a
    list of gamma ray energies.

    Keyword arguments::
        eng_gams: List of energies of photon in laboratory frame.
        eng_k: Energy of charged kaon in laboratory frame.
    """
    return CSpectrum(eng_gams, eng_k, mode)



import numpy as np

from hazma.decay import muon
from hazma.decay import neutral_pion, charged_pion

from hazma.parameters import neutral_pion_mass as mpi0

from hazma.vector_mediator.vector_mediator_decay_spectrum import (
    dnde_decay_v,
    dnde_decay_v_pt,
)


class VectorMediatorSpectra:
    def dnde_ee(self, e_gams, e_cm, spectrum_type="all"):
        fsr = np.vectorize(self.dnde_xx_to_v_to_ffg)

        if spectrum_type == "all":
            return self.dnde_ee(e_gams, e_cm, "fsr") + self.dnde_ee(
                e_gams, e_cm, "decay"
            )
        elif spectrum_type == "fsr":
            return fsr(e_gams, e_cm, "e")
        elif spectrum_type == "decay":
            return np.array([0.0 for _ in range(len(e_gams))])
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def dnde_mumu(self, e_gams, e_cm, spectrum_type="all"):
        fsr = np.vectorize(self.dnde_xx_to_v_to_ffg)  # todo: this line
        decay = np.vectorize(muon)

        if spectrum_type == "all":
            return self.dnde_mumu(e_gams, e_cm, "fsr") + self.dnde_mumu(
                e_gams, e_cm, "decay"
            )
        elif spectrum_type == "fsr":
            return fsr(e_gams, e_cm, "mu")
        elif spectrum_type == "decay":
            return 2.0 * decay(e_gams, e_cm / 2.0)
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def dnde_pi0g(self, e_gams, e_cm, spectrum_type="all"):
        if spectrum_type == "all":
            return self.dnde_pi0g(e_gams, e_cm, "fsr") + self.dnde_pi0g(
                e_gams, e_cm, "decay"
            )
        elif spectrum_type == "fsr":
            return np.array([0.0 for _ in range(len(e_gams))])
        elif spectrum_type == "decay":
            # Neutral pion's energy
            e_pi0 = (e_cm ** 2 + mpi0 ** 2) / (2.0 * e_cm)

            return neutral_pion(e_gams, e_pi0)
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def dnde_pipi(self, e_gams, e_cm, spectrum_type="all"):
        if spectrum_type == "all":
            return self.dnde_pipi(e_gams, e_cm, "fsr") + self.dnde_pipi(
                e_gams, e_cm, "decay"
            )
        elif spectrum_type == "fsr":
            return self.dnde_xx_to_v_to_pipig(e_gams, e_cm)
        elif spectrum_type == "decay":
            return 2.0 * charged_pion(e_gams, e_cm / 2.0)
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def __dnde_v(self, e_gams, e_v, fs="total"):
        """
        Helper function for computing the spectrum from V decaying with
        arbitrary boost.
        """
        mv = self.mv
        pws = self.partial_widths()
        pw_array = np.zeros(5, dtype=float)

        # Check is the decay width of the vector is zero.
        if pws["total"] == 0.0:
            if hasattr(e_gams, "__len__"):
                return np.array([0.0] * len(e_gams))
            return 0.0

        pw_array[0] = pws["e e"] / pws["total"]
        pw_array[1] = pws["mu mu"] / pws["total"]
        pw_array[2] = pws["pi0 g"] / pws["total"]
        pw_array[3] = pws["pi pi"] / pws["total"]

        if hasattr(e_gams, "__len__"):
            return dnde_decay_v(e_gams, e_v, mv, pw_array, fs)

        return dnde_decay_v_pt(e_gams, e_v, mv, pw_array, fs)

    def dnde_pi0v(self, e_gams, e_cm):
        e_pi0 = (e_cm ** 2 + mpi0 ** 2 - self.mv ** 2) / (2 * e_cm)
        pi0_spec = neutral_pion(e_gams, e_pi0)

        e_v = (e_cm ** 2 - mpi0 ** 2 + self.mv ** 2) / (2 * e_cm)
        v_spec = self.__dnde_v(e_gams, e_v, fs="total")

        return pi0_spec + v_spec

    def dnde_vv(self, e_gams, e_cm, fs="total"):
        # Each vector gets half the COM energy
        return 2.0 * self.__dnde_v(e_gams, e_cm / 2.0, fs)

    def spectrum_funcs(self):
        """
        Returns a dictionary of all the avaiable spectrum functions for
        a pair of initial state fermions with mass `mx` annihilating into
        each available final state.

        Each argument of the spectrum functions in `eng_gams`, an array
        of the gamma ray energies to evaluate the spectra at and `e_cm`, the
        center of mass energy of the process.

        Note
        ----
        This does not return a function for computing the spectrum for the pi0
        pi pi final state since it always contributes orders of magnitude less
        than the pi pi and pi0 g final states.
        """
        return {
            "mu mu": self.dnde_mumu,
            "e e": self.dnde_ee,
            "pi pi": self.dnde_pipi,
            "pi0 g": self.dnde_pi0g,
            "pi0 v": self.dnde_pi0v,
            "v v": self.dnde_vv,
        }

    def gamma_ray_lines(self, e_cm):
        bf = self.annihilation_branching_fractions(e_cm)["pi0 g"]

        return {"pi0 g": {"energy": (e_cm ** 2 - mpi0 ** 2) / (2.0 * e_cm), "bf": bf}}



import numpy as np
cimport numpy as np
import cython
from libc.math cimport sqrt

include "parameters.pxd"

from scipy.integrate import quad

cdef double mmu = MASS_MU
cdef double me = MASS_E

@cython.cdivision(True)
cdef double __spectrum_rf(double eng_p):
    """
    Returns the positron spectrum from a muon in the muon rest frame.

    Parameters
    ----------
    eng_p : float
        Energy of the positron.

    Returns
    -------
    dnde : float
        The value of the spectrum given a positron energy `eng_p`.
    """
    cdef double r = me / mmu
    cdef double s = me * me - 2. * eng_p * mmu + mmu * mmu
    cdef double smax = (mmu - me) * (mmu - me)
    cdef double smin = 0.0
    if s <= smin or smax <= s:
        return 0.0

    return 2 * mmu * (2 * (pow(mmu, 4) * pow(-1 + r * r, 2) + mmu * mmu *
                           (1 + r * r) * s - 2 * s * s) *
                      np.sqrt(pow(mmu, 4) * pow(-1 + r * r, 2) -
                              2 * mmu**2 * (1 + r * r) * s + s * s)) / pow(mmu, 8)

@cython.cdivision(True)
cdef double __integrand(double cl, double eng_p, double eng_mu):
    """
    Returns the integrand for the boost integral used to compute the postitron
    spectrum from the muon.

    Parameters
    ----------
    cl : double
        Angle the electron makes with the z-axis.
    eng_p : double
        Energy of the positron.
    eng_mu : double
        Energy of the muon.

    Returns
    -------
    integrand : double
        Integral for the boost integral.
    """
    if eng_p < me:
        return 0.0
    cdef double p = sqrt(eng_p * eng_p - me * me)
    cdef double gamma = eng_mu / mmu
    cdef double beta = sqrt(1.0 - pow(mmu / eng_mu, 2))
    cdef double emurf = gamma * (eng_p - p * beta * cl)
    cdef double jac = p / (2. * sqrt((1 + pow(beta * cl, 2)) * eng_p * eng_p -
                                     (1 + beta * beta * (-1 + cl * cl)) *
                                     me * me -
                                     2 * beta * cl * eng_p * p) * gamma)

    return __spectrum_rf(emurf) * jac

@cython.cdivision(True)
cdef double CSpectrumPoint(double eng_p, double eng_mu):
    """
    Returns the positron spectrum at a single electron energy from the muon
    given an arbitrary muon energy.

    Parameters
    ----------
    eng_p : double
        Energy of the positron.
    eng_mu : double
        Energy of the muon.

    Returns
    -------
    dnde : double
        Positron spectrum of the muon given an electron energy `eng_p` and a muon
        energy `eng_mu`.
    """
    if eng_mu < mmu:
        return 0.0

    return quad(__integrand, -1., 1., points=[-1.0, 1.0],
                args=(eng_p, eng_mu), epsabs=1e-10, epsrel=1e-4)[0]

@cython.boundscheck(True)
@cython.wraparound(False)
cdef np.ndarray CSpectrum(np.ndarray engs_p, double eng_mu):
    """
    Returns the positron spectrum at many electron energies from the muon
    given an arbitrary muon energy.

    Parameters
    ----------
    engs_p : numpy.array
        Energy of the positron.
    eng_mu : double
        Energy of the muon.

    Returns
    -------
    dnde : numpy.array
        Positron spectrum of the muon given electron energies `engs_p` and a
        muon energy `eng_mu`.
    """
    cdef int numpts = len(engs_p)

    cdef np.ndarray spec = np.zeros(numpts, dtype=np.float64)

    cdef int i = 0

    for i in range(numpts):
        spec[i] = CSpectrumPoint(engs_p[i], eng_mu)

    return spec

def SpectrumPoint(double eng_p, double eng_mu):
    """
    Returns the positron spectrum at a single electron energy from the muon
    given an arbitrary muon energy.

    Parameters
    ----------
    eng_p : double
        Energy of the positron.
    eng_mu : double
        Energy of the muon.

    Returns
    -------
    dnde : double
        Positron spectrum of the muon given an electron energy `eng_p` and a muon
        energy `eng_mu`.
    """
    return CSpectrumPoint(eng_p, eng_mu)

@cython.boundscheck(True)
@cython.wraparound(False)
def Spectrum(np.ndarray engs_p, double eng_mu):
    """
    Returns the positron spectrum at many electron energies from the muon
    given an arbitrary muon energy.

    Parameters
    ----------
    engs_p : numpy.array
        Energy of the positron.
    eng_mu : double
        Energy of the muon.

    Returns
    -------
    dnde : numpy.array
        Positron spectrum of the muon given electron energies `engs_p` and a
        muon energy `eng_mu`.
    """
    return CSpectrum(engs_p, eng_mu)



import numpy as np
cimport numpy as np

cdef np.ndarray __e_gams_total
cdef np.ndarray __e_gams_000
cdef np.ndarray __e_gams_penu
cdef np.ndarray __e_gams_penug
cdef np.ndarray __e_gams_pm0
cdef np.ndarray __e_gams_pm0g
cdef np.ndarray __e_gams_pmunu
cdef np.ndarray __e_gams_pmunug

cdef np.ndarray __spec_total
cdef np.ndarray __spec_000
cdef np.ndarray __spec_penu
cdef np.ndarray __spec_penug
cdef np.ndarray __spec_pm0
cdef np.ndarray __spec_pm0g
cdef np.ndarray __spec_pmunu
cdef np.ndarray __spec_pmunug

cdef double __interp_spec(double, str)
cdef double __integrand(double, double, double, str)

cdef double CSpectrumPoint(double, double, str)
cdef np.ndarray CSpectrum(np.ndarray[np.float64_t, ndim=1], double, str)



"""
# Module for computing relativistic phase space points.

# Authors: Logan Morrison and Adam Coogan
# Date: December 2017

TODO: Code up specific functions for cross-section
      functions for 2->2 processes.
TODO: Code up specific functions for cross-section
      functions for 2->3 processes.

"""
from hazma.phase_space_helper_functions import generator
from hazma.phase_space_helper_functions import histogram
from hazma.phase_space_helper_functions.modifiers import apply_matrix_elem
import numpy as np
import multiprocessing as mp
import warnings

from hazma.hazma_errors import RamboCMETooSmall

from hazma.field_theory_helper_functions.common_functions import cross_section_prefactor


def generate_phase_space_point(masses, cme):
    """
    Generate a phase space point given a set of final state particles and a
    given center of mass energy.

    Parameters
    ----------
    masses : numpy.ndarray
        List of masses of the final state particles.
    cme : double
        Center-of-mass-energy of the process.

    Returns
    -------
    phase_space_points : numpy.ndarray
        List of four momenta and a event weight. The returned numpy array is of
        the form::

            [ke1, kx1, ky1, kz1, ..., keN, kxN, kyN, kzN, weight]

    """

    if not hasattr(masses, "__len__"):
        masses = [masses]

    masses = np.array(masses)
    return generator.generate_point(masses, cme, len(masses))


def generate_phase_space(
    masses, cme, num_ps_pts=10000, mat_elem_sqrd=lambda klist: 1, num_cpus=None
):
    """
    Generate a specified number of phase space points given a set of
    final state particles and a given center of mass energy.
    Note that the weights are not normalized.

    Parameters
    ----------
    masses : numpy.ndarray
        List of masses of the final state particles.
    cme : double
        Center-of-mass-energy of the process.
    num_ps_pts : int {10000]
        Total number of phase space points to generate.
    mat_elem_sqrd : (double)(numpy.ndarray) {lambda klist: 1]
        Function for the matrix element squared.
    num_cpus : int {None]
        Number of cpus to use in parallel with rambo. If not specified, 75% of
        the cpus will be used.

    Returns
    -------
    phase_space_points : numpy.ndarray
        List of phase space points. The phase space points are in the form::

            [[ke11, kx11, ky11, kz11, ..., keN1, kxN1, kyN1, kzN1, weight1],
             ...
             [ke1N, kx1N, ky1N, kz1N, ..., keNN, kxNN, kyNN, kzNN, weightN]]

    Examples
    --------

    Generate 100000 phase space points for a 3 body final state::

        from hazma import rambo
        import numpy as np
        masses = np.array([100., 200., 0.0])
        cme = 10.0 * sum(masses)
        num_ps_pts = 100000
        rambo.generate_phase_space(masses, cme, num_ps_pts=num_ps_pts)

    """
    if not hasattr(masses, "__len__"):
        masses = [masses]

    if cme < sum(masses):
        raise RamboCMETooSmall()

    num_fsp = len(masses)
    # If the user doesn't specify the number of cpus to use,
    # use 75% of them.
    if num_cpus is not None:
        if num_cpus > num_ps_pts:
            num_cpus = num_ps_pts
        if num_cpus > mp.cpu_count():
            num_cpus = int(np.floor(mp.cpu_count() * 0.75))
            warnings.warn(
                """You only have {] cpus.
                          Using {] cpus instead.
                          """.format(
                    mp.cpu_count(), num_cpus
                )
            )
    if num_cpus is None:
        # Use 75% of the cpu power.
        num_cpus = int(np.floor(mp.cpu_count() * 0.75))
        # If user wants a number of phase space points which is less
        # than the number of cpus available, use num_ps_pts cpus instead.
        if num_cpus > num_ps_pts:
            num_cpus = num_ps_pts
    # Instantiate `num_cpus` number of workers and divide num_ps_pts among the
    # the workers to speed up phase space generation.
    pool = mp.Pool(num_cpus)
    num_ps_pts_per_cpu = int(num_ps_pts / num_cpus)
    # If num_ps_pts % num_cpus !=0, then we need to compute the actual number
    # of phase space points.
    actual_num_ps_pts = num_ps_pts_per_cpu * num_cpus
    # Create a container to store the results from the workers
    job_results = []
    # Run the jobs on 75% of the cpus.
    for i in range(num_cpus):
        job_results.append(
            pool.apply_async(
                generator.generate_space, (num_ps_pts_per_cpu, masses, cme, num_fsp)
            )
        )
    # Close the pool and wait for results to finish
    pool.close()
    pool.join()
    # Put results in a numpy array.
    points = np.array([result.get() for result in job_results])
    # Flatten the outer axis to have a list of phase space points.
    points = points.reshape(actual_num_ps_pts, 4 * num_fsp + 1)
    # Resize the weights to have the correct cross section.
    points = apply_matrix_elem(points, actual_num_ps_pts, num_fsp, mat_elem_sqrd)

    return points


def generate_energy_histogram(
    masses,
    cme,
    num_ps_pts=10000,
    mat_elem_sqrd=lambda klist: 1,
    num_bins=25,
    num_cpus=None,
    density=False,
):
    """
    Generate energy histograms for each of the final state particles.

    Parameters
    ----------
    num_ps_pts : int
        Total number of phase space points to generate.
    masses : numpy.ndarray
        List of masses of the final state particles.
    cme : double
        Center-of-mass-energy of the process.
    mat_elem_sqrd : (double)(numpy.ndarray) {lambda klist: 1]
        Function for the matrix element squared.
    num_bins : int
        Number of energy bins to use for each of the final state particles.
    num_cpus : int {None]
        Number of cpus to use in parallel with rambo. If not specified, 75% of
        the cpus will be used.
    density: Bool
        If true, the histograms will be normalized to have unit area underneath
        the curves, i.e. they will be probability density functions.

    Returns
    -------
    energy_histograms : numpy.ndarray
        List of energies and dsigma/dE's. The resulting array has the shape
        (num_fsp, 2, num_bins). The array is formatted as::

            [[[E11, E12, ....], [hist11, hist12, ...]],
             ...
             [[EN1, EN2, ....], [histM1, histN2, ...]]]

    Examples
    --------

    Making energy histograms for 4 final state particles and plotting their
    energy spectra::

        from hazma import rambo
        import numpy as np
        num_ps_pts = 100000
        masses = np.array([100., 100., 0.0, 0.0])
        cme = 1000.
        num_bins = 100
        eng_hist = rambo.generate_energy_histogram(masses, cme,
                                                   num_ps_pts=num_ps_pts
                                                   num_bins=num_bins)
        import matplotlib.pyplot as plt
        for i in range(len(masses)):
            plt.loglog(eng_hist[i, 0], eng_hist[i, 1])

    """
    if not hasattr(masses, "__len__"):
        masses = [masses]

    if cme < sum(masses):
        raise RamboCMETooSmall()

    num_fsp = len(masses)

    pts = generate_phase_space(masses, cme, num_ps_pts, mat_elem_sqrd, num_cpus)

    actual_num_ps_pts = pts.shape[0]

    return histogram.space_to_energy_hist(
        pts, actual_num_ps_pts, num_fsp, num_bins, density=density
    )


def integrate_over_phase_space(
    fsp_masses, cme, num_ps_pts=10000, mat_elem_sqrd=lambda momenta: 1, num_cpus=None
):
    """
    Returns the integral over phase space given a squared matrix element, a
    set of final state particle masses and a given energy.

    Parameters
    ----------
    fsp_masses : numpy.ndarray
        List of masses of the final state particles.
    cme : double
        Center-of-mass-energy of the process.
    num_ps_pts : int {10000]
        Total number of phase space points to generate.
    mat_elem_sqrd : (double)(numpy.ndarray) {lambda momenta: 1]
        Function for the matrix element squared.
    num_cpus : int {None]
        Number of cpus to use in parallel with rambo. If not specified, 75% of
        the cpus will be used.

    Returns
    -------
    integral : float
        The result of the integral over phase space.
    std : float
        The estimated error in the integral over phase space.

    """
    if not hasattr(fsp_masses, "__len__"):
        fsp_masses = [fsp_masses]

    if cme < sum(fsp_masses):
        raise RamboCMETooSmall()

    num_fsp = len(fsp_masses)
    points = generate_phase_space(fsp_masses, cme, num_ps_pts, mat_elem_sqrd, num_cpus)
    actual_num_ps_pts = len(points[:, 4 * num_fsp])
    weights = points[:, 4 * num_fsp]
    integral = np.average(weights)
    std = np.std(weights) / np.sqrt(actual_num_ps_pts)

    return integral, std


def compute_annihilation_cross_section(
    isp_masses,
    fsp_masses,
    cme,
    num_ps_pts=10000,
    mat_elem_sqrd=lambda momenta: 1,
    num_cpus=None,
):
    """
    Computes the cross section for a given process.

    Parameters
    ----------
    isp_masses : numpy.ndarray
        List of masses of the initial state particles.
    fsp_masses : numpy.ndarray
        List of masses of the final state and particles.
    cme : double
        Center-of-mass-energy of the process.
    num_ps_pts : int
        Total number of phase space points to generate.
    mat_elem_sqrd : (double)(numpy.ndarray) {lambda momenta: 1]
        Function for the matrix element squared.
    num_cpus : int {None]
        Number of cpus to use in parallel with rambo. If not specified, 75% of
        the cpus will be used.

    Returns
    -------
    cross_section : double
        Cross section for X -> final state particles(fsp), where the fsp have
        masses `masses` and the process X -> fsp has a squared matrix element
        of `mat_elem_sqrd`.
    std : double
        Estimated error in cross section.

    Examples
    --------

    Compute the cross section for electrons annihilating into muons through a
    photon. First, we construct a function for the matrix element::

        from hazma.parameters import electron_mass as me
        from hazma.parameters import muon_mass as mmu
        from hazma.parameters import qe
        MDot = lambda p1, p2: (p1[0] * p2[0] - p1[1] * p1[1] - p1[2] * p1[2] -
                               p1[3] * p1[3])
        def msqrd(momenta):
            Q = sum(momenta)[0] # center-of-mass energy
            # Momenta of the incoming electrons
            p1 = np.array([Q / 2., 0., 0., np.sqrt(Q**2 / 4. - me**2)])
            p2 = np.array([Q / 2., 0., 0., -np.sqrt(Q**2 / 4. - me**2)])
            # Momenta for the outgoing muons
            p3 = momenta[0]
            p4 = momenta[1]
            # Mandelstam variables
            s = MDot(p1 + p2, p1 + p2)
            t = MDot(p1 - p3, p1 - p3)
            u = MDot(p1 - p4, p1 - p4)
            return (2 * qe**4 * (t**2 + u**2 - 4 * (t + u)* me**2 +
                    6 * me**4 + 4 * s * mmu**2 + 2 * mmu**4)) / s**2

    Next we integrate over phase space using RAMBO::

        from hazma.rambo import compute_annihilation_cross_section
        import numpy as np
        isp_masses = np.array([me, me])
        fsp_masses = np.array([mmu, mmu])
        cme = 1000.
        compute_annihilation_cross_section(
            isp_masses, fsp_masses, cme, num_ps_pts=5000, mat_elem_sqrd=msqrd)

    """
    integral, std = integrate_over_phase_space(
        fsp_masses,
        cme,
        num_ps_pts=num_ps_pts,
        mat_elem_sqrd=mat_elem_sqrd,
        num_cpus=num_cpus,
    )

    m1 = isp_masses[0]
    m2 = isp_masses[1]

    cross_section = integral * cross_section_prefactor(m1, m2, cme)
    error = cross_section_prefactor(m1, m2, cme) * std

    return cross_section, error


def compute_decay_width(
    fsp_masses, cme, num_ps_pts=10000, mat_elem_sqrd=lambda momenta: 1, num_cpus=None
):
    r"""
    Computes the decay width for a given process.

    Parameters
    ----------
    fsp_masses : numpy.ndarray
        List of masses of the final state particles.
    cme : double
        Center-of-mass-energy of the process.
    num_ps_pts : int
        Total number of phase space points to generate.
    mat_elem_sqrd : (double)(numpy.ndarray) {lambda momenta: 1]
        Function for the matrix element squared.
    num_cpus : int {None]
        Number of cpus to use in parallel with rambo. If not specified, 75% of
        the cpus will be used.

    Returns
    -------
    cross_section : double
        Cross section for X -> final state particles (FSPs), where the FSPs have
        masses `masses` and the process X -> FSPs has a squared matrix element
        of `mat_elem_sqrd`.
    std : double
        Estimated error in cross section.

    Examples
    --------

    In this example we compute the partial decay width of the muon for
    :math:`\mu~\to~e\nu\nu`

    First, we declare the matrix element::

        from hazma.parameters import GF
        MDot = lambda p1, p2: (p1[0] * p2[0] - p1[1] * p1[1] - p1[2] * p1[2] -
                               p1[3] * p1[3])
        def msqrd(momenta):
            pe = momenta[0]
            pve = momenta[1]
            pvmu = momenta[2]
            pmu = sum(momenta)
            return 64. * GF**2 * MDot(pe, pvmu) * MDot(pmu, pve)

    Next, we compute the decay width::

        from hazma.rambo import compute_decay_width
        from hazma.parameters import electron_mass as me
        from hazma.parameters import muon_mass as mmu
        import numpy as np
        fsp_masses = np.array([me, 0.0, 0.0])
        cme = mmu
        compute_decay_width(fsp_masses, cme, mat_elem_sqrd=msqrd)
    """
    integral, std = integrate_over_phase_space(
        fsp_masses,
        cme,
        num_ps_pts=num_ps_pts,
        mat_elem_sqrd=mat_elem_sqrd,
        num_cpus=num_cpus,
    )

    cross_section = integral / (2.0 * cme)
    error = std / (2.0 * cme)

    return cross_section, error



import numpy as np

from hazma.parameters import vh, b0, alpha_em, fpi
from hazma.parameters import muon_mass as mmu
from hazma.parameters import electron_mass as me
from hazma.parameters import up_quark_mass as muq
from hazma.parameters import down_quark_mass as mdq
from hazma.parameters import charged_pion_mass as mpi

from scipy.integrate import quad


class PseudoScalarMediatorWidths:
    def width_p_to_gg(self):
        """Returns the partial decay width of the pseudoscalar decaying into
        photons.
        """
        beta = self.beta
        gpFF = self.gpFF
        mp = self.mp

        ret = -(
            alpha_em ** 2
            * mp ** 3
            * ((1 + beta) * fpi * gpFF - beta * vh)
            * ((-1 + beta) * fpi * gpFF + beta * vh)
        ) / (128.0 * fpi ** 2 * np.pi ** 3 * vh ** 2)

        assert ret.imag == 0
        assert ret.real >= 0

        return ret.real

    def width_p_to_xx(self):
        mp = self.mp
        rx = self.mx / mp

        if 2.0 * rx < 1:
            ret = -(
                (-1 + self.beta ** 2) * self.gpxx ** 2 * mp * np.sqrt(1 - 4 * rx ** 2)
            ) / (32.0 * np.pi)

            assert ret.imag == 0
            assert ret.real >= 0

            return ret.real
        else:
            return 0.0

    def width_p_to_ff(self, f):
        mp = self.mp

        if f == "e":
            rf = me / mp
            gpff = self.gpee
        elif f == "mu":
            rf = mmu / mp
            gpff = self.gpmumu

        if 2.0 * rf < 1:
            ret = -(
                (-1 + self.beta ** 2) * gpff ** 2 * mp * np.sqrt(1 - 4 * rf ** 2)
            ) / (8.0 * np.pi)

            assert ret.imag == 0
            assert ret.real >= 0

            return ret.real
        else:
            return 0.0

    def dwidth_ds_p_to_pi0pi0pi0(self, s):
        mp = self.mp
        mpi0 = self.mpi0  # use shifted pion mass!

        if mp >= 3.0 * mpi0:
            gpuu = self.gpuu
            gpdd = self.gpdd
            gpGG = self.gpGG
            beta = self.beta

            ret = -(
                b0 ** 2
                * np.sqrt(s * (-4 * mpi0 ** 2 + s))
                * np.sqrt(
                    mp ** 4 + (mpi0 ** 2 - s) ** 2 - 2 * mp ** 2 * (mpi0 ** 2 + s)
                )
                * (
                    -(beta ** 2 * (mdq + muq) ** 2 * vh ** 2)
                    + 2
                    * beta
                    * fpi
                    * (mdq + muq)
                    * vh
                    * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh)
                    + (-1 + 10 * beta ** 2)
                    * fpi ** 2
                    * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh) ** 2
                )
            ) / (256.0 * fpi ** 4 * mp ** 3 * np.pi ** 3 * s * vh ** 2)

            assert ret.imag == 0
            assert ret.real >= 0

            return ret
        else:
            return 0.0

    def width_p_to_pi0pi0pi0(self):
        """
        Returns the width for the pseudoscalar's decay into three neutral
        pions.

        Parameters
        ----------
        self : PseudoScalarMediator or PseudoScalarMediatorParameters object
            Object containing the parameters of the pseudo-scalar mediator
            model. Can be a PseudoScalarMediator or a
            PseudoScalarMediatorParameters object.

        Returns
        -------
        gamma : float
            The width for P -> pi0 pi0 pi0.
        """
        mp = self.mp
        mpi0 = self.mpi0  # use shifted pion mass!

        smax = (mp - mpi0) ** 2
        smin = 4.0 * mpi0 ** 2

        res = quad(self.dwidth_ds_p_to_pi0pi0pi0, smin, smax)

        return res[0]

    def dwidth_ds_p_to_pi0pipi(self, s):
        mp = self.mp
        mpi0 = self.mpi0  # use shifted pion mass!

        if mp >= 2.0 * mpi + mpi0:
            gpuu = self.gpuu
            gpdd = self.gpdd
            gpGG = self.gpGG
            beta = self.beta

            ret = (
                np.sqrt(s * (-4 * mpi ** 2 + s))
                * np.sqrt(
                    mp ** 4 + (mpi0 ** 2 - s) ** 2 - 2 * mp ** 2 * (mpi0 ** 2 + s)
                )
                * (
                    beta ** 2 * (2 * mpi ** 2 + mpi0 - 3 * s) ** 2 * vh ** 2
                    + 2
                    * b0
                    * beta
                    * (2 * mpi ** 2 + mpi0 - 3 * s)
                    * vh
                    * (
                        -(beta * (mdq + muq) * vh)
                        + fpi * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh)
                    )
                    + b0 ** 2
                    * (
                        beta ** 2 * (mdq + muq) ** 2 * vh ** 2
                        - 2
                        * beta
                        * fpi
                        * (mdq + muq)
                        * vh
                        * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh)
                        - (-1 + 4 * beta ** 2)
                        * fpi ** 2
                        * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh) ** 2
                    )
                )
            ) / (2304.0 * fpi ** 4 * mp ** 3 * np.pi ** 3 * s * vh ** 2)

            ret = (
                np.sqrt(s * (-4 * mpi ** 2 + s))
                * np.sqrt(
                    mp ** 4 + (mpi0 ** 2 - s) ** 2 - 2 * mp ** 2 * (mpi0 ** 2 + s)
                )
                * (
                    beta ** 2 * (2 * mpi ** 2 + mpi0 ** 2 - 3 * s) ** 2 * vh ** 2
                    + 2
                    * b0
                    * beta
                    * (2 * mpi ** 2 + mpi0 ** 2 - 3 * s)
                    * vh
                    * (
                        -(beta * (mdq + muq) * vh)
                        + fpi * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh)
                    )
                    + b0 ** 2
                    * (
                        beta ** 2 * (mdq + muq) ** 2 * vh ** 2
                        - 2
                        * beta
                        * fpi
                        * (mdq + muq)
                        * vh
                        * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh)
                        - (-1 + 4 * beta ** 2)
                        * fpi ** 2
                        * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh) ** 2
                    )
                )
            ) / (2304.0 * fpi ** 4 * mp ** 3 * np.pi ** 3 * s * vh ** 2)

            assert ret.imag == 0
            assert ret.real >= 0

            return ret
        else:
            return 0.0

    def width_p_to_pi0pipi(self):
        """
        Returns the width for the pseudoscalar's decay into a neutral pion and
        two charged pions.

        Parameters
        ----------
        self : PseudoScalarMediator or PseudoScalarMediatorParameters object
            Object containing the parameters of the pseudo-scalar mediator
            model. Can be a PseudoScalarMediator or a
            PseudoScalarMediatorParameters object.

        Returns
        -------
        gamma : float
            The width for P -> pi0 pi+ pi-.
        """
        mp = self.mp
        mpi0 = self.mpi0  # use shifted pion mass!

        smax = (mp - mpi0) ** 2
        smin = 4.0 * mpi ** 2

        res = quad(self.dwidth_ds_p_to_pi0pipi, smin, smax)

        return res[0]

    def partial_widths(self):
        """
        Returns a dictionary for the partial decay widths of the pseudoscalar
        mediator.

        Returns
        -------
        width_dict : dictionary
            Dictionary of all of the individual decay widths of the
            pseudoscalar mediator as well as the total decay width.
        """
        w_gg = self.width_p_to_gg()
        w_xx = self.width_p_to_xx()

        w_ee = self.width_p_to_ff("e")
        w_mumu = self.width_p_to_ff("mu")

        w_pi0pipi = self.width_p_to_pi0pipi()
        w_pi0pi0pi0 = self.width_p_to_pi0pi0pi0()

        total = w_gg + w_xx + w_ee + w_mumu + w_pi0pipi + w_pi0pi0pi0

        width_dict = {
            "g g": w_gg,
            "x x": w_xx,
            "e e": w_ee,
            "mu mu": w_mumu,
            "pi0 pi pi": w_pi0pipi,
            "pi0 pi0 pi0": w_pi0pi0pi0,
            "total": total,
        }

        return width_dict



# from cmath import sqrt, pi, log
from hazma.parameters import charged_pion_mass as mpi
from hazma.parameters import neutral_pion_mass as mpi0

# from ..parameters import fpi
# from ..parameters import qe
from hazma.parameters import muon_mass as mmu
from hazma.parameters import electron_mass as me
from scipy.integrate import quad


class AxialVectorMediatorCrossSections:
    def sigma_xx_to_a_to_ff(self, Q, f):
        """
        Returns the cross section for xbar x to fbar f.

        Parameters
        ----------
        Q : float
            Center of mass energy.
        f : float
            Name of final state fermion: "e" or "mu".
        self : object
            Class containing the vector mediator parameters.

        Returns
        -------
        cross_section : float
            Cross section for xbar + x -> a -> fbar + f.
        """
        if f == "e":
            mf = me
            # gall = self.gaee
        elif f == "mu":
            mf = mmu
            # gall = self.gamumu
        mx = self.mx
        if Q >= 2.0 * mf and Q >= 2.0 * mx:
            # gaxx = self.gaxx
            # ma = self.ma
            # width_a = self.width_a
            ret_val = 0.0
            assert ret_val.imag == 0
            assert ret_val.real >= 0
            return ret_val.real
        else:
            return 0.0

    def dsigma_ds_xx_to_a_to_pi0pipi(self, s, Q):
        mx = self.mx

        if (
            Q > 2.0 * mpi + mpi0
            and Q > 2.0 * mx
            and s > 4.0 * mpi ** 2
            and s < (Q - mpi0) ** 2
        ):
            # gauu = self.gauu
            # gadd = self.gadd
            # gaxx = self.gaxx
            # ma = self.ma
            # width_a = self.width_a

            ret_val = 0.0

            assert ret_val.imag == 0.0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_a_to_pi0pipi(self, Q):
        if Q > 2.0 * mpi + mpi0 and Q > 2.0 * self.mx:
            s_min = 4.0 * mpi ** 2
            s_max = (Q - mpi0) ** 2

            ret_val = quad(self.dsigma_ds_xx_to_a_to_pi0pipi, s_min, s_max, args=(Q))[0]

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_aa(self, Q):
        mx = self.mx
        ma = self.ma

        if Q >= 2.0 * ma and Q >= 2.0 * mx:
            # gaxx = self.gaxx

            ret_val = 0.0

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def annihilation_cross_sections(self, Q):
        """
        Compute the total cross section for two fermions annihilating through a
        vector mediator to mesons and leptons.

        Parameters
        ----------
        cme : float
            Center of mass energy.

        Returns
        -------
        cs : float
            Total cross section.
        """
        muon_contr = self.sigma_xx_to_a_to_ff(Q, "mu")
        electron_contr = self.sigma_xx_to_a_to_ff(Q, "e")
        pi0pipi_contr = self.sigma_xx_to_a_to_pi0pipi(Q)
        aa_contr = self.sigma_xx_to_aa(Q)

        total = muon_contr + electron_contr + pi0pipi_contr + aa_contr
        # pi0pipi_contr

        cross_secs = {
            "mu mu": muon_contr,
            "e e": electron_contr,
            "pi0 pi pi": pi0pipi_contr,
            "a a": aa_contr,
            "total": total,
        }

        return cross_secs

    def annihilation_branching_fractions(self, Q):
        """
        Compute the branching fractions for two fermions annihilating through
        an axial vector mediator to mesons and leptons.

        Parameters
        ----------
        Q : float
            Center of mass energy.

        Returns
        -------
        bfs : dictionary
            Dictionary of the branching fractions. The keys are 'total',
            'mu mu', 'e e', 'pi0 pi pi', 'a a'.
        """
        CSs = self.cross_sections(Q)

        if CSs["total"] == 0.0:
            return {"mu mu": 0.0, "e e": 0.0, "pi0 pi pi": 0.0, "a a": 0.0}
        else:
            return {
                "mu mu": CSs["mu mu"] / CSs["total"],
                "e e": CSs["e e"] / CSs["total"],
                "pi0 pi pi": CSs["pi0 pi pi"] / CSs["total"],
                "a a": CSs["a a"] / CSs["total"],
            }
